from contextlib import ExitStack
from typing import Mapping, MutableSet

from afancontrol.config import FanName
from afancontrol.logger import logger
from afancontrol.pwmfan import PWMFanNorm, PWMValueNorm
from afancontrol.report import Report


class Fans:
    def __init__(self, fans: Mapping[FanName, PWMFanNorm], *, report: Report) -> None:
        self.fans = fans
        self.report = report
        self._stack = None

        # Set of fans marked as failing (which speed is 0)
        self._failed_fans = set()  # type: MutableSet[FanName]

        # Set of fans that will be skipped on speed check
        self._stopped_fans = set()  # type: MutableSet[FanName]

    def is_fan_failing(self, fan_name: FanName) -> bool:
        return fan_name in self._failed_fans

    def is_fan_stopped(self, fan_name: FanName) -> bool:
        return fan_name in self._stopped_fans

    def __enter__(self):  # reusable
        self._stack = ExitStack()
        logger.info("Enabling PWM on fans...")
        try:
            for pwmfan in self.fans.values():
                self._stack.enter_context(pwmfan)
        except Exception:
            self._stack.close()
            raise
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        logger.info("Disabling PWM on fans...")
        self._stack.close()
        logger.info("Done. Fans should be returned to full speed")
        return None

    def check_speeds(self) -> None:
        for name, fan in self.fans.items():
            if name in self._stopped_fans:
                continue
            try:
                if fan.get_speed() <= 0:
                    raise RuntimeError("Fan speed is 0")
            except Exception as e:
                self._ensure_fan_is_failing(name, e)
            else:
                self._ensure_fan_is_not_failing(name)

    def set_all_to_full_speed(self) -> None:
        for name, fan in self.fans.items():
            if name in self._failed_fans:
                continue
            try:
                fan.set_full_speed()
            except Exception as e:
                logger.warning("Unable to set the fan '%s' to full speed:\n%s", name, e)

    def set_fan_speeds(self, speeds: Mapping[FanName, PWMValueNorm]) -> None:
        assert speeds.keys() == self.fans.keys()
        self._stopped_fans.clear()
        for name, pwm_norm in speeds.items():
            fan = self.fans[name]
            assert 0.0 <= pwm_norm <= 1.0

            if name in self._failed_fans:
                return

            try:
                pwm = fan.set(pwm_norm)
            except Exception as e:
                logger.warning(
                    "Unable to set the fan '%s' to speed %s:\n%s", name, pwm_norm, e
                )
            else:
                logger.debug(
                    "Fan status [%s]: speed: %.3f, pwm: %s", name, pwm_norm, pwm
                )
                if fan.is_pwm_stopped(pwm):
                    self._stopped_fans.add(name)

    def _ensure_fan_is_failing(self, name: FanName, get_speed_exc: Exception) -> None:
        if name in self._failed_fans:
            return
        self._failed_fans.add(name)
        fan = self.fans[name]
        try:
            # Perhaps it had jammed, so setting it to full speed might
            # recover it?
            fan.set_full_speed()
        except Exception as e:
            full_speed_result = "Setting fan speed to full has failed:\n%s" % e
        else:
            full_speed_result = "Fan has been set to full speed"

        self.report.report(
            "fan stopped: %s" % name,
            "Looks like the fan '%s' is failing:\n%s\n\n%s"
            % (name, get_speed_exc, full_speed_result),
        )

    def _ensure_fan_is_not_failing(self, name: FanName) -> None:
        if name not in self._failed_fans:
            return
        self.report.report(
            "fan started: %s" % name,
            "Fan '%s' which had previously been reported as failing has just started."
            % name,
        )
        self._failed_fans.remove(name)



import abc
from contextlib import ExitStack
from typing import Mapping, Optional, Set

from afancontrol.config import AlertCommands, TempName, TriggerConfig
from afancontrol.exec import exec_shell_command
from afancontrol.logger import logger
from afancontrol.report import Report
from afancontrol.temp import TempStatus


class Trigger(abc.ABC):
    def __init__(
        self,
        *,
        global_commands: AlertCommands,
        temp_commands: Mapping[TempName, AlertCommands],
        report: Report
    ) -> None:
        self.global_commands = global_commands
        self.temp_commands = temp_commands
        self.report = report
        self._alerting_temps = set()  # type: Set[TempName]

    @property
    @abc.abstractmethod
    def trigger_name(self):
        pass

    def __enter__(self):  # reusable
        assert not self._alerting_temps
        self._alerting_temps.clear()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        if self.is_alerting:
            # Although the exceptional situation is not yet resolved,
            # we call the corresponding leave callbacks, because
            # if this is a reload, then we might get the enter callbacks
            # being executed twice.
            # Perhaps we should somehow tell the `leave_cmd` that it's
            # being called not because the issue is resolved,
            # but because this program is restarting or is shutting down.
            self.report.report(
                "Leaving %s MODE" % self.trigger_name.upper(),
                "Leaving %s MODE because of shutting down or restarting."
                % self.trigger_name.upper(),
            )
            for name in self._alerting_temps:
                self._alert_cmd(self.temp_commands[name].leave_cmd)
            self._alert_cmd(self.global_commands.leave_cmd)

        self._alerting_temps.clear()
        return None

    @property
    def is_alerting(self) -> bool:
        return bool(self._alerting_temps)

    def check(self, temps: Mapping[TempName, Optional[TempStatus]]) -> None:
        was_alerting = self.is_alerting
        self._update_alerting_temps(temps)
        self._process_global_alerting_commands(temps, was_alerting, self.is_alerting)

    def _update_alerting_temps(
        self, temps: Mapping[TempName, Optional[TempStatus]]
    ) -> None:
        stopped_alerting_temps = self._alerting_temps.copy()
        for name, status in temps.items():
            temp_alerting_reason = self._temp_alerting_reason(status)
            if not temp_alerting_reason:
                continue
            if name in self._alerting_temps:
                # Still alerting
                stopped_alerting_temps.discard(name)
                continue

            # Just started alerting
            self._alerting_temps.add(name)
            logger.warning(
                "%s started on temp. name: %s, status: %s, reason: %s",
                self.trigger_name.upper(),
                name,
                status,
                temp_alerting_reason,
            )
            self._alert_cmd(self.temp_commands[name].enter_cmd)

        for name in stopped_alerting_temps:
            self._alerting_temps.discard(name)
            status = temps[name]

            logger.warning(
                "%s ended on temp: name: %s, status: %s",
                self.trigger_name.upper(),
                name,
                status,
            )
            self._alert_cmd(self.temp_commands[name].leave_cmd)

    def _process_global_alerting_commands(
        self,
        temps: Mapping[TempName, Optional[TempStatus]],
        was_alerting: bool,
        is_alerting: bool,
    ) -> None:
        is_entered = not was_alerting and is_alerting
        is_left = was_alerting and not is_alerting
        if is_entered or is_left:
            temps_debug = "\n".join(
                "[%s]: %s" % (name, status)
                for name, status in sorted(temps.items(), key=lambda kv: kv[0])
            )
            if is_entered:
                self.report.report(
                    "Entered %s MODE" % self.trigger_name.upper(),
                    "Entered %s MODE. Take a look as soon as possible!!!\nSensors:\n%s"
                    % (self.trigger_name.upper(), temps_debug),
                )
                self._alert_cmd(self.global_commands.enter_cmd)
            if is_left:
                self.report.report(
                    "Leaving %s MODE" % self.trigger_name.upper(),
                    "Leaving %s MODE.\nSensors:\n%s"
                    % (self.trigger_name.upper(), temps_debug),
                )
                self._alert_cmd(self.global_commands.leave_cmd)

    @abc.abstractmethod
    def _temp_alerting_reason(self, temp: Optional[TempStatus]) -> Optional[str]:
        pass

    def _alert_cmd(self, shell_cmd):
        if not shell_cmd:
            return
        try:
            exec_shell_command(shell_cmd)
        except Exception as e:
            logger.warning(
                "Enable to execute %s trigger command %s:\n%s",
                self.trigger_name,
                shell_cmd,
                e,
            )


class PanicTrigger(Trigger):
    trigger_name = "panic"

    def _temp_alerting_reason(self, temp: Optional[TempStatus]) -> Optional[str]:
        if temp is None:
            return "Sensor failed"
        if not temp.is_panic:
            return None
        return "Panic temp reached"


class ThresholdTrigger(Trigger):
    trigger_name = "threshold"

    def _temp_alerting_reason(self, temp: Optional[TempStatus]) -> Optional[str]:
        if temp is None:
            return None
        if not temp.is_threshold:
            return None
        return "Threshold temp reached"


class Triggers:
    def __init__(self, triggers_config: TriggerConfig, report: Report) -> None:
        self.panic_trigger = PanicTrigger(
            global_commands=triggers_config.global_commands.panic,
            temp_commands={
                temp_name: actions.panic
                for temp_name, actions in triggers_config.temp_commands.items()
            },
            report=report,
        )
        self.threshold_trigger = ThresholdTrigger(
            global_commands=triggers_config.global_commands.threshold,
            temp_commands={
                temp_name: actions.threshold
                for temp_name, actions in triggers_config.temp_commands.items()
            },
            report=report,
        )
        self._stack = None

    def __enter__(self):  # reusable
        self._stack = ExitStack()
        try:
            self._stack.enter_context(self.panic_trigger)
            self._stack.enter_context(self.threshold_trigger)
        except Exception:
            self._stack.close()
            raise
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._stack.close()
        return None

    @property
    def is_alerting(self) -> bool:
        return self.panic_trigger.is_alerting or self.threshold_trigger.is_alerting

    def check(self, temps: Mapping[TempName, Optional[TempStatus]]) -> None:
        self.panic_trigger.check(temps)
        self.threshold_trigger.check(temps)



import json
import socket
import threading
import traceback
from contextlib import ExitStack
from time import sleep
from typing import Dict

import pytest

from afancontrol.arduino import (
    ArduinoConnection,
    ArduinoName,
    ArduinoPin,
    ArduinoPWMFan,
    SetPWMCommand,
    pyserial_available,
)
from afancontrol.pwmfan import PWMValue

pytestmark = pytest.mark.skipif(
    not pyserial_available, reason="pyserial is not installed"
)


class DummyArduino:
    """Emulate an Arduino board, i.e. the other side of the pyserial connection.

    Slightly mimics the Arduino program `micro.ino`.
    """

    def __init__(self) -> None:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(("127.0.0.1", 0))
        s.listen(1)
        listening_port = s.getsockname()[1]
        self.sock = s
        self.pyserial_url = "socket://127.0.0.1:%s" % listening_port
        self._lock = threading.Lock()
        self._loop_iteration_complete = threading.Event()
        self._first_loop_iteration_complete = threading.Event()
        self._disconnected = threading.Event()
        self._thread_error = threading.Event()
        self._is_connected = False
        self._inner_state_pwms = {"5": 255, "9": 255, "10": 255, "11": 255}
        self._inner_state_speeds = {"0": 0, "1": 0, "2": 0, "3": 0, "7": 0}

    def set_inner_state_pwms(self, pwms: Dict[str, int]) -> None:
        with self._lock:
            self._inner_state_pwms.update(pwms)
        if self.is_connected:
            self._loop_iteration_complete.clear()
            assert self._loop_iteration_complete.wait(5) is True

    def set_speeds(self, speeds: Dict[str, int]) -> None:
        with self._lock:
            self._inner_state_speeds.update(speeds)
        if self.is_connected:
            self._loop_iteration_complete.clear()
            assert self._loop_iteration_complete.wait(5) is True

    @property
    def inner_state_pwms(self):
        with self._lock:
            copy = self._inner_state_pwms.copy()
        return copy

    @property
    def is_connected(self):
        with self._lock:
            if not self._is_connected:
                return False
        assert self._first_loop_iteration_complete.wait(5) is True
        return True

    def wait_for_disconnected(self):
        assert self._disconnected.wait(5) is True

    def accept(self):
        client, _ = self.sock.accept()
        self.sock.close()  # Don't accept any more connections
        with self._lock:
            self._is_connected = True
        threading.Thread(target=self._thread_run, args=(client,), daemon=True).start()

    def _thread_run(self, sock):
        sock.settimeout(0.001)
        command_buffer = bytearray()
        try:
            while True:
                # The code in this loop mimics the `loop` function
                # in the `micro.ino` program.

                try:
                    command_buffer.extend(sock.recv(1024))
                except socket.timeout:
                    pass

                while len(command_buffer) >= 3:
                    command_raw = command_buffer[:3]
                    del command_buffer[:3]
                    command = SetPWMCommand.parse(command_raw)
                    with self._lock:
                        self._inner_state_pwms[str(command.pwm_pin)] = command.pwm

                sock.sendall(self._make_status())

                self._loop_iteration_complete.set()
                self._first_loop_iteration_complete.set()

                sleep(0.050)
        except (ConnectionResetError, BrokenPipeError):
            pass
        except Exception:
            traceback.print_exc()
            self._thread_error.set()
        finally:
            with self._lock:
                self._is_connected = False
            sock.close()
            self._disconnected.set()

    def _make_status(self):
        with self._lock:
            status = {
                "fan_inputs": self._inner_state_speeds,
                "fan_pwm": self._inner_state_pwms,
            }
            return (json.dumps(status) + "\n").encode("ascii")

    def ensure_no_errors_in_thread(self):
        assert self._thread_error.is_set() is not True


@pytest.fixture
def dummy_arduino():
    return DummyArduino()


def test_smoke(dummy_arduino):
    conn = ArduinoConnection(ArduinoName("test"), dummy_arduino.pyserial_url)
    fan = ArduinoPWMFan(conn, pwm_pin=ArduinoPin(9), tacho_pin=ArduinoPin(3))

    dummy_arduino.set_inner_state_pwms({"9": 42})

    with ExitStack() as stack:
        assert not dummy_arduino.is_connected
        stack.enter_context(fan)
        dummy_arduino.accept()
        assert dummy_arduino.is_connected

        dummy_arduino.set_speeds({"3": 1200})
        conn.wait_for_status()  # required only for synchronization in the tests
        assert fan.get_speed() == 1200
        assert fan.get() == 255
        assert dummy_arduino.inner_state_pwms["9"] == 255

        fan.set(PWMValue(192))
        dummy_arduino.set_speeds({"3": 998})
        conn.wait_for_status()  # required only for synchronization in the tests
        assert fan.get_speed() == 998
        assert fan.get() == 192
        assert dummy_arduino.inner_state_pwms["9"] == 192

    dummy_arduino.wait_for_disconnected()
    assert dummy_arduino.inner_state_pwms["9"] == 255
    assert not dummy_arduino.is_connected
    dummy_arduino.ensure_no_errors_in_thread()



from unittest.mock import call

from afancontrol import report
from afancontrol.report import Report


def test_report_success(sense_exec_shell_command):
    r = Report(r"printf '@%s' '%REASON%' '%MESSAGE%'")

    with sense_exec_shell_command(report) as (mock_exec_shell_command, get_stdout):
        r.report("reason here", "message\nthere")
        assert mock_exec_shell_command.call_args == call(
            "printf '@%s' 'reason here' 'message\nthere'"
        )
        assert ["@reason here@message\nthere"] == get_stdout()


def test_report_fail_does_not_raise():
    r = Report("false")
    r.report("reason here", "message\nthere")



import tempfile
from contextlib import contextmanager
from pathlib import Path
from unittest.mock import patch

import pytest

from afancontrol.exec import exec_shell_command


@pytest.fixture
def temp_path():
    with tempfile.TemporaryDirectory() as tmpdirname:
        yield Path(tmpdirname).resolve()


@pytest.fixture
def sense_exec_shell_command():
    exec_shell_command_stdout = []

    def sensed_exec_shell_command(*args, **kwargs):
        exec_shell_command_stdout.append(exec_shell_command(*args, **kwargs))
        return exec_shell_command_stdout[-1]

    def get_stdout():
        try:
            return exec_shell_command_stdout[:]
        finally:
            exec_shell_command_stdout.clear()

    @contextmanager
    def _sense_exec_shell_command(module):
        with patch.object(
            module, "exec_shell_command", wraps=sensed_exec_shell_command
        ) as mock_exec_shell_command:
            yield mock_exec_shell_command, get_stdout

    return _sense_exec_shell_command



from pathlib import Path
from unittest.mock import Mock

import pytest

from afancontrol.arduino import (
    ArduinoConnection,
    ArduinoName,
    ArduinoPin,
    ArduinoPWMFan,
    pyserial_available,
)
from afancontrol.config import (
    Actions,
    AlertCommands,
    DaemonCLIConfig,
    DaemonConfig,
    FanName,
    FanSpeedModifier,
    FansTempsRelation,
    MappingName,
    ParsedConfig,
    TempName,
    TriggerConfig,
    parse_config,
)
from afancontrol.pwmfan import (
    FanInputDevice,
    LinuxPWMFan,
    PWMDevice,
    PWMFanNorm,
    PWMValue,
)
from afancontrol.temp import FileTemp, HDDTemp, TempCelsius


@pytest.fixture
def example_conf():
    return Path(__file__).parents[1] / "afancontrol.conf"


def path_from_str(contents: str) -> Path:
    p = Mock(spec=Path)
    p.read_text.return_value = contents
    return p


@pytest.mark.skipif(not pyserial_available, reason="pyserial is not installed")
def test_example_conf(example_conf: Path):
    daemon_cli_config = DaemonCLIConfig(
        pidfile=None, logfile=None, exporter_listen_host=None
    )

    parsed = parse_config(example_conf, daemon_cli_config)
    assert parsed == ParsedConfig(
        daemon=DaemonConfig(
            pidfile="/var/run/afancontrol.pid",
            logfile="/var/log/afancontrol.log",
            exporter_listen_host=None,
            interval=5,
        ),
        report_cmd=(
            'printf "Subject: %s\nTo: %s\n\n%b" '
            '"afancontrol daemon report: %REASON%" root "%MESSAGE%" | sendmail -t'
        ),
        triggers=TriggerConfig(
            global_commands=Actions(
                panic=AlertCommands(enter_cmd=None, leave_cmd=None),
                threshold=AlertCommands(enter_cmd=None, leave_cmd=None),
            ),
            temp_commands={
                TempName("hdds"): Actions(
                    panic=AlertCommands(enter_cmd=None, leave_cmd=None),
                    threshold=AlertCommands(enter_cmd=None, leave_cmd=None),
                ),
                TempName("mobo"): Actions(
                    panic=AlertCommands(enter_cmd=None, leave_cmd=None),
                    threshold=AlertCommands(enter_cmd=None, leave_cmd=None),
                ),
            },
        ),
        fans={
            FanName("cpu"): PWMFanNorm(
                LinuxPWMFan(
                    PWMDevice("/sys/class/hwmon/hwmon0/device/pwm1"),
                    FanInputDevice("/sys/class/hwmon/hwmon0/device/fan1_input"),
                ),
                pwm_line_start=PWMValue(100),
                pwm_line_end=PWMValue(240),
                never_stop=True,
            ),
            FanName("hdd"): PWMFanNorm(
                LinuxPWMFan(
                    PWMDevice("/sys/class/hwmon/hwmon0/device/pwm2"),
                    FanInputDevice("/sys/class/hwmon/hwmon0/device/fan2_input"),
                ),
                pwm_line_start=PWMValue(100),
                pwm_line_end=PWMValue(240),
                never_stop=False,
            ),
            FanName("my_arduino_fan"): PWMFanNorm(
                ArduinoPWMFan(
                    ArduinoConnection(
                        ArduinoName("mymicro"),
                        "/dev/ttyACM0",  # linux
                        # "/dev/cu.usbmodem14201",  # macos
                        baudrate=115200,
                        status_ttl=5,
                    ),
                    pwm_pin=ArduinoPin(9),
                    tacho_pin=ArduinoPin(3),
                ),
                pwm_line_start=PWMValue(100),
                pwm_line_end=PWMValue(240),
                never_stop=True,
            ),
        },
        temps={
            TempName("hdds"): HDDTemp(
                "/dev/sd?",
                min=TempCelsius(35.0),
                max=TempCelsius(48.0),
                panic=TempCelsius(55.0),
                threshold=None,
                hddtemp_bin="hddtemp",
            ),
            TempName("mobo"): FileTemp(
                "/sys/class/hwmon/hwmon0/device/temp1_input",
                min=TempCelsius(30.0),
                max=TempCelsius(40.0),
                panic=None,
                threshold=None,
            ),
        },
        mappings={
            MappingName("1"): FansTempsRelation(
                temps=[TempName("mobo"), TempName("hdds")],
                fans=[
                    FanSpeedModifier(fan=FanName("cpu"), modifier=1.0),
                    FanSpeedModifier(fan=FanName("hdd"), modifier=0.6),
                    FanSpeedModifier(fan=FanName("my_arduino_fan"), modifier=1.0),
                ],
            ),
            MappingName("2"): FansTempsRelation(
                temps=[TempName("hdds")],
                fans=[FanSpeedModifier(fan=FanName("hdd"), modifier=1.0)],
            ),
        },
    )


def test_minimal_config() -> None:
    daemon_cli_config = DaemonCLIConfig(
        pidfile=None, logfile=None, exporter_listen_host=None
    )

    config = """
[daemon]

[actions]

[temp:mobo]
type = file
path = /sys/class/hwmon/hwmon0/device/temp1_input

[fan: case]
pwm = /sys/class/hwmon/hwmon0/device/pwm2
fan_input = /sys/class/hwmon/hwmon0/device/fan2_input

[mapping:1]
fans = case*0.6,
temps = mobo
"""
    parsed = parse_config(path_from_str(config), daemon_cli_config)
    assert parsed == ParsedConfig(
        daemon=DaemonConfig(
            pidfile="/var/run/afancontrol.pid",
            logfile=None,
            exporter_listen_host=None,
            interval=5,
        ),
        report_cmd=(
            'printf "Subject: %s\nTo: %s\n\n%b" '
            '"afancontrol daemon report: %REASON%" root "%MESSAGE%" | sendmail -t'
        ),
        triggers=TriggerConfig(
            global_commands=Actions(
                panic=AlertCommands(enter_cmd=None, leave_cmd=None),
                threshold=AlertCommands(enter_cmd=None, leave_cmd=None),
            ),
            temp_commands={
                TempName("mobo"): Actions(
                    panic=AlertCommands(enter_cmd=None, leave_cmd=None),
                    threshold=AlertCommands(enter_cmd=None, leave_cmd=None),
                )
            },
        ),
        fans={
            FanName("case"): PWMFanNorm(
                LinuxPWMFan(
                    PWMDevice("/sys/class/hwmon/hwmon0/device/pwm2"),
                    FanInputDevice("/sys/class/hwmon/hwmon0/device/fan2_input"),
                ),
                pwm_line_start=PWMValue(100),
                pwm_line_end=PWMValue(240),
                never_stop=True,
            )
        },
        temps={
            TempName("mobo"): FileTemp(
                "/sys/class/hwmon/hwmon0/device/temp1_input",
                min=None,
                max=None,
                panic=None,
                threshold=None,
            )
        },
        mappings={
            MappingName("1"): FansTempsRelation(
                temps=[TempName("mobo")],
                fans=[FanSpeedModifier(fan=FanName("case"), modifier=0.6)],
            )
        },
    )



import abc
import re
from pathlib import Path
from typing import NamedTuple, NewType, Optional, Tuple

from afancontrol.exec import exec_shell_command

TempCelsius = NewType("TempCelsius", float)

TempStatus = NamedTuple(
    "TempStatus",
    [
        ("temp", TempCelsius),
        ("min", TempCelsius),
        ("max", TempCelsius),
        ("panic", Optional[TempCelsius]),
        ("threshold", Optional[TempCelsius]),
        ("is_panic", bool),
        ("is_threshold", bool),
    ],
)


class Temp(abc.ABC):
    def __init__(
        self, *, panic: Optional[TempCelsius], threshold: Optional[TempCelsius]
    ) -> None:
        self._panic = panic
        self._threshold = threshold

    def get(self) -> TempStatus:
        temp, min_t, max_t = self._get_temp()

        if not (min_t < max_t):
            raise RuntimeError(
                "Min temperature must be less than max. %s < %s" % (min_t, max_t)
            )

        return TempStatus(
            temp=temp,
            min=min_t,
            max=max_t,
            panic=self._panic,
            threshold=self._threshold,
            is_panic=self._panic is not None and temp >= self._panic,
            is_threshold=self._threshold is not None and temp >= self._threshold,
        )

    @abc.abstractmethod
    def _get_temp(self) -> Tuple[TempCelsius, TempCelsius, TempCelsius]:
        pass


class FileTemp(Temp):
    def __init__(
        self,
        temp_path: str,  # /sys/class/hwmon/hwmon0/temp1
        *,
        min: Optional[TempCelsius],
        max: Optional[TempCelsius],
        panic: Optional[TempCelsius],
        threshold: Optional[TempCelsius]
    ) -> None:
        super().__init__(panic=panic, threshold=threshold)
        temp_path = re.sub(r"_input$", "", temp_path)
        self._temp_input = Path(temp_path + "_input")
        self._temp_min = Path(temp_path + "_min")
        self._temp_max = Path(temp_path + "_max")
        self._min = min
        self._max = max

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (
                self._temp_input == other._temp_input
                and self._temp_min == other._temp_min
                and self._temp_max == other._temp_max
                and self._min == other._min
                and self._max == other._max
                and self._panic == other._panic
                and self._threshold == other._threshold
            )

        return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s(%r, min=%r, max=%r, panic=%r, threshold=%r)" % (
            type(self).__name__,
            str(self._temp_input),
            self._min,
            self._max,
            self._panic,
            self._threshold,
        )

    def _get_temp(self) -> Tuple[TempCelsius, TempCelsius, TempCelsius]:
        temp = self._read_temp_from_path(self._temp_input)
        return temp, self._get_min(), self._get_max()

    def _get_min(self) -> TempCelsius:
        if self._min is not None:
            return self._min
        try:
            min_t = self._read_temp_from_path(self._temp_min)
        except FileNotFoundError:
            raise RuntimeError(
                "Please specify `min` and `max` temperatures for "
                "the %s sensor" % self._temp_input
            )
        return min_t

    def _get_max(self) -> TempCelsius:
        if self._max is not None:
            return self._max
        try:
            max_t = self._read_temp_from_path(self._temp_max)
        except FileNotFoundError:
            raise RuntimeError(
                "Please specify `min` and `max` temperatures for "
                "the %s sensor" % self._temp_input
            )
        return max_t

    @staticmethod
    def _read_temp_from_path(path: Path) -> TempCelsius:
        return TempCelsius(int(path.read_text().strip()) / 1000)


class HDDTemp(Temp):
    def __init__(
        self,
        disk_path: str,
        *,
        min: TempCelsius,
        max: TempCelsius,
        panic: Optional[TempCelsius],
        threshold: Optional[TempCelsius],
        hddtemp_bin: str = "hddtemp"
    ) -> None:
        super().__init__(panic=panic, threshold=threshold)
        self._disk_path = disk_path
        self._min = min
        self._max = max
        self._hddtemp_bin = hddtemp_bin

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (
                self._disk_path == other._disk_path
                and self._min == other._min
                and self._max == other._max
                and self._panic == other._panic
                and self._threshold == other._threshold
                and self._hddtemp_bin == other._hddtemp_bin
            )

        return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s(%r, min=%r, max=%r, panic=%r, threshold=%r, hddtemp_bin=%r)" % (
            type(self).__name__,
            self._disk_path,
            self._min,
            self._max,
            self._panic,
            self._threshold,
            self._hddtemp_bin,
        )

    def _get_temp(self) -> Tuple[TempCelsius, TempCelsius, TempCelsius]:
        temps = [
            float(line.strip())
            for line in self._call_hddtemp().split("\n")
            if self._is_float(line.strip())
        ]
        if not temps:
            raise RuntimeError(
                "hddtemp returned empty list of valid temperature values"
            )
        temp = TempCelsius(max(temps))
        return temp, self._get_min(), self._get_max()

    def _get_min(self) -> TempCelsius:
        return TempCelsius(self._min)

    def _get_max(self) -> TempCelsius:
        return TempCelsius(self._max)

    def _call_hddtemp(self) -> str:
        # `disk_path` might be a glob, so it has to be executed with a shell.
        shell_command = "%s -n -u C -- %s" % (self._hddtemp_bin, self._disk_path)
        return exec_shell_command(shell_command)

    @staticmethod
    def _is_float(s: str) -> bool:
        if not s:
            return False
        try:
            float(s)
        except (ValueError, TypeError):
            return False
        else:
            return True


class CommandTemp(Temp):
    def __init__(
        self,
        shell_command: str,
        *,
        min: Optional[TempCelsius],
        max: Optional[TempCelsius],
        panic: Optional[TempCelsius],
        threshold: Optional[TempCelsius]
    ) -> None:
        super().__init__(panic=panic, threshold=threshold)
        self._shell_command = shell_command
        self._min = min
        self._max = max

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (
                self._shell_command == other._shell_command
                and self._min == other._min
                and self._max == other._max
                and self._panic == other._panic
                and self._threshold == other._threshold
            )

        return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s(%r, min=%r, max=%r, panic=%r, threshold=%r)" % (
            type(self).__name__,
            self._shell_command,
            self._min,
            self._max,
            self._panic,
            self._threshold,
        )

    def _get_temp(self) -> Tuple[TempCelsius, TempCelsius, TempCelsius]:
        temps = [
            float(line.strip())
            for line in exec_shell_command(self._shell_command).split("\n")
            if line.strip()
        ]
        temp = TempCelsius(temps[0])

        if self._min is not None:
            min_t = self._min
        else:
            min_t = TempCelsius(temps[1])

        if self._max is not None:
            max_t = self._max
        else:
            max_t = TempCelsius(temps[2])

        return temp, min_t, max_t



from contextlib import ExitStack
from typing import Type
from unittest.mock import MagicMock, patch

import pytest
from click.testing import CliRunner

from afancontrol import fantest
from afancontrol.fantest import (
    CSVMeasurementsOutput,
    HumanMeasurementsOutput,
    MeasurementsOutput,
    fantest as main,
    run_fantest,
)
from afancontrol.pwmfan import (
    BasePWMFan,
    FanInputDevice,
    LinuxPWMFan,
    PWMDevice,
    PWMValue,
)


def test_main_smoke(temp_path):
    pwm_path = temp_path / "pwm2"
    pwm_path.write_text("")
    fan_input_path = temp_path / "fan2_input"
    fan_input_path.write_text("")

    with ExitStack() as stack:
        mocked_fantest = stack.enter_context(patch.object(fantest, "run_fantest"))

        runner = CliRunner()
        result = runner.invoke(
            main,
            [
                "--fan-type",
                "linux",
                "--linux-fan-pwm",
                # "/sys/class/hwmon/hwmon0/device/pwm2",
                str(pwm_path),  # click verifies that this file exists
                "--linux-fan-input",
                # "/sys/class/hwmon/hwmon0/device/fan2_input",
                str(fan_input_path),  # click verifies that this file exists
                "--output-format",
                "human",
                "--direction",
                "increase",
                "--pwm-step-size",
                "accurate",
            ],
        )

        print(result.output)
        assert result.exit_code == 0

        assert mocked_fantest.call_count == 1

        args, kwargs = mocked_fantest.call_args
        assert not args
        assert kwargs.keys() == {"fan", "pwm_step_size", "output"}
        assert kwargs["fan"] == LinuxPWMFan(
            PWMDevice(str(pwm_path)), FanInputDevice(str(fan_input_path))
        )
        assert kwargs["pwm_step_size"] == 5
        assert isinstance(kwargs["output"], HumanMeasurementsOutput)


@pytest.mark.parametrize("pwm_step_size", [5, -5])
@pytest.mark.parametrize("output_cls", [HumanMeasurementsOutput, CSVMeasurementsOutput])
def test_fantest(output_cls: Type[MeasurementsOutput], pwm_step_size: PWMValue):
    mocked_fan = MagicMock(spec=BasePWMFan)
    mocked_fan.min_pwm = 0
    mocked_fan.max_pwm = 255
    output = output_cls()

    with ExitStack() as stack:
        mocked_sleep = stack.enter_context(patch.object(fantest, "sleep"))
        mocked_fan.get_speed.return_value = 999

        run_fantest(fan=mocked_fan, pwm_step_size=pwm_step_size, output=output)

        assert mocked_fan.set.call_count == (255 // abs(pwm_step_size)) + 1
        assert mocked_fan.get_speed.call_count == (255 // abs(pwm_step_size))
        assert mocked_sleep.call_count == (255 // abs(pwm_step_size)) + 1

        if pwm_step_size > 0:
            # increase
            expected_set = [0] + list(range(0, 255, pwm_step_size))
        else:
            # decrease
            expected_set = [255] + list(range(255, 0, pwm_step_size))
        assert [pwm for (pwm,), _ in mocked_fan.set.call_args_list] == expected_set



import abc
import math
from pathlib import Path
from typing import NewType

PWMDevice = NewType("PWMDevice", str)
FanInputDevice = NewType("FanInputDevice", str)
PWMValue = NewType("PWMValue", int)  # [0..255]
PWMValueNorm = NewType("PWMValueNorm", float)  # [0..1]
FanValue = NewType("FanValue", int)


class BasePWMFan(abc.ABC):
    max_pwm = PWMValue(255)
    min_pwm = PWMValue(0)

    def is_stopped(self) -> bool:
        return type(self).is_pwm_stopped(self.get())

    @staticmethod
    def is_pwm_stopped(pwm: PWMValue) -> bool:
        return pwm <= 0

    @abc.abstractmethod
    def get(self) -> PWMValue:
        pass

    def set(self, pwm: PWMValue) -> None:
        if not (type(self).min_pwm <= pwm <= type(self).max_pwm):
            raise ValueError(
                "Invalid pwm value %s: it must be within [%s..%s]"
                % (pwm, type(self).min_pwm, type(self).max_pwm)
            )
        self._set_raw(pwm)

    @abc.abstractmethod
    def _set_raw(self, pwm: PWMValue) -> None:
        pass

    def set_full_speed(self) -> None:
        self._set_raw(type(self).max_pwm)

    @abc.abstractmethod
    def get_speed(self) -> FanValue:
        pass

    def __enter__(self):  # reusable
        """Enable PWM control for this fan"""
        self._enable_pwm()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Disable PWM control for this fan"""
        self._disable_pwm()

    @abc.abstractmethod
    def _enable_pwm(self) -> None:
        pass

    @abc.abstractmethod
    def _disable_pwm(self) -> None:
        pass


class LinuxPWMFan(BasePWMFan):
    def __init__(self, pwm: PWMDevice, fan_input: FanInputDevice) -> None:
        super().__init__()
        self._pwm = Path(pwm)
        self._pwm_enable = Path(pwm + "_enable")
        self._fan_input = Path(fan_input)

    def get(self) -> PWMValue:
        return PWMValue(int(self._pwm.read_text()))

    def _set_raw(self, pwm: PWMValue) -> None:
        self._pwm.write_text(str(int(pwm)))

    def get_speed(self) -> FanValue:
        return FanValue(int(self._fan_input.read_text()))

    def _enable_pwm(self) -> None:
        # fancontrol way of doing it
        if self._pwm_enable.is_file():
            self._pwm_enable.write_text("1")
        self.set_full_speed()

    def _disable_pwm(self) -> None:
        # fancontrol way of doing it
        if not self._pwm_enable.is_file():
            self.set_full_speed()
            return

        self._pwm_enable.write_text("0")
        if self._pwm_enable.read_text() == "0":
            return

        self._pwm_enable.write_text("1")
        self.set_full_speed()

        if self._pwm_enable.read_text() == "1" and self.get() >= type(self).max_pwm:
            return

        raise RuntimeError("Couldn't disable PWM on the fan %r" % self)

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (
                self._pwm == other._pwm
                and self._pwm_enable == other._pwm_enable
                and self._fan_input == other._fan_input
            )

        return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s(%r, %r)" % (
            type(self).__name__,
            str(self._pwm),
            str(self._fan_input),
        )


class PWMFanNorm:
    def __init__(
        self,
        pwmfan: BasePWMFan,
        *,
        pwm_line_start: PWMValue,
        pwm_line_end: PWMValue,
        never_stop: bool = False
    ) -> None:
        self.pwmfan = pwmfan
        self.pwm_line_start = pwm_line_start
        self.pwm_line_end = pwm_line_end
        self.never_stop = never_stop
        if type(self.pwmfan).min_pwm > self.pwm_line_start:
            raise ValueError(
                "Invalid pwm_line_start. Expected: min_pwm <= pwm_line_start. "
                "Got: %s <= %s" % (type(self.pwmfan).min_pwm, self.pwm_line_start)
            )
        if self.pwm_line_end > type(self.pwmfan).max_pwm:
            raise ValueError(
                "Invalid pwm_line_end. Expected: pwm_line_end <= max_pwm. "
                "Got: %s <= %s" % (self.pwm_line_end, type(self.pwmfan).max_pwm)
            )

    def __enter__(self):
        self.pwmfan.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        return self.pwmfan.__exit__(exc_type, exc_value, exc_tb)

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (
                self.pwmfan == other.pwmfan
                and self.pwm_line_start == other.pwm_line_start
                and self.pwm_line_end == other.pwm_line_end
                and self.never_stop == other.never_stop
            )

        return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s(%r, pwm_line_start=%r, pwm_line_end=%r, never_stop=%r)" % (
            type(self).__name__,
            self.pwmfan,
            self.pwm_line_start,
            self.pwm_line_end,
            self.never_stop,
        )

    def is_pwm_stopped(self, pwm: PWMValue) -> bool:
        return type(self.pwmfan).is_pwm_stopped(pwm)

    def set_full_speed(self) -> None:
        self.pwmfan.set_full_speed()

    def get_speed(self) -> FanValue:
        return self.pwmfan.get_speed()

    def get_raw(self) -> PWMValue:
        return self.pwmfan.get()

    def get(self) -> PWMValueNorm:
        return PWMValueNorm(self.get_raw() / self.pwmfan.max_pwm)

    def set(self, pwm_norm: PWMValueNorm) -> PWMValue:
        # TODO validate this formula
        pwm_norm = max(pwm_norm, PWMValueNorm(0.0))
        pwm_norm = min(pwm_norm, PWMValueNorm(1.0))
        pwm = pwm_norm * self.pwm_line_end
        if 0 < pwm < self.pwm_line_start:
            pwm = self.pwm_line_start
        if pwm <= 0 and self.never_stop:
            pwm = self.pwm_line_start
        if pwm_norm >= 1.0:
            pwm = self.pwmfan.max_pwm

        pwm = PWMValue(int(math.ceil(pwm)))
        self.pwmfan.set(pwm)
        return pwm



from unittest.mock import MagicMock

import pytest

from afancontrol.config import FanName
from afancontrol.fans import Fans
from afancontrol.pwmfan import BasePWMFan, PWMFanNorm, PWMValueNorm
from afancontrol.report import Report


@pytest.fixture
def report():
    return MagicMock(spec=Report)


@pytest.mark.parametrize("is_fan_failing", [False, True])
def test_smoke(report, is_fan_failing):
    fan = MagicMock(spec=PWMFanNorm)
    fans = Fans({FanName("test"): fan}, report=report)

    fan.set = lambda pwm_norm: int(255 * pwm_norm)
    fan.get_speed.return_value = 0 if is_fan_failing else 942
    fan.is_pwm_stopped = BasePWMFan.is_pwm_stopped

    with fans:
        assert 1 == fan.__enter__.call_count
        fans.check_speeds()
        fans.set_all_to_full_speed()
        fans.set_fan_speeds({FanName("test"): PWMValueNorm(0.42)})
        assert fan.get_speed.call_count == 1
        if is_fan_failing:
            assert fans._failed_fans == {"test"}
            assert fans._stopped_fans == set()
        else:
            assert fans._failed_fans == set()
            assert fans._stopped_fans == set()

    assert 1 == fan.__exit__.call_count



from contextlib import ExitStack
from unittest.mock import MagicMock, patch

import pytest

import afancontrol.manager
from afancontrol.config import (
    Actions,
    AlertCommands,
    FanName,
    FanSpeedModifier,
    FansTempsRelation,
    MappingName,
    TempName,
    TriggerConfig,
)
from afancontrol.manager import Manager
from afancontrol.metrics import Metrics
from afancontrol.pwmfan import PWMFanNorm
from afancontrol.report import Report
from afancontrol.temp import FileTemp
from afancontrol.trigger import Triggers


@pytest.fixture
def report():
    return MagicMock(spec=Report)


def test_manager(report):
    mocked_case_fan = MagicMock(spec=PWMFanNorm)()
    mocked_mobo_temp = MagicMock(spec=FileTemp)()
    mocked_metrics = MagicMock(spec=Metrics)()

    with ExitStack() as stack:
        stack.enter_context(
            patch.object(afancontrol.manager, "Triggers", spec=Triggers)
        )

        manager = Manager(
            fans={FanName("case"): mocked_case_fan},
            temps={TempName("mobo"): mocked_mobo_temp},
            mappings={
                MappingName("1"): FansTempsRelation(
                    temps=[TempName("mobo")],
                    fans=[FanSpeedModifier(fan=FanName("case"), modifier=0.6)],
                )
            },
            report=report,
            triggers_config=TriggerConfig(
                global_commands=Actions(
                    panic=AlertCommands(enter_cmd=None, leave_cmd=None),
                    threshold=AlertCommands(enter_cmd=None, leave_cmd=None),
                ),
                temp_commands={
                    TempName("mobo"): Actions(
                        panic=AlertCommands(enter_cmd=None, leave_cmd=None),
                        threshold=AlertCommands(enter_cmd=None, leave_cmd=None),
                    )
                },
            ),
            metrics=mocked_metrics,
        )

        stack.enter_context(manager)

        manager.tick()

        mocked_triggers = manager.triggers  # type: MagicMock
        assert mocked_triggers.check.call_count == 1
        assert mocked_case_fan.__enter__.call_count == 1
        assert mocked_metrics.__enter__.call_count == 1
        assert mocked_metrics.tick.call_count == 1
    assert mocked_case_fan.__exit__.call_count == 1
    assert mocked_metrics.__exit__.call_count == 1



import configparser
from pathlib import Path
from typing import (
    Dict,
    Mapping,
    NamedTuple,
    NewType,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
)

from afancontrol.arduino import (
    DEFAULT_BAUDRATE,
    DEFAULT_STATUS_TTL,
    ArduinoConnection,
    ArduinoName,
    ArduinoPin,
    ArduinoPWMFan,
)
from afancontrol.pwmfan import (
    BasePWMFan,
    FanInputDevice,
    LinuxPWMFan,
    PWMDevice,
    PWMFanNorm,
    PWMValue,
)
from afancontrol.temp import CommandTemp, FileTemp, HDDTemp, Temp, TempCelsius

DEFAULT_CONFIG = "/etc/afancontrol/afancontrol.conf"
DEFAULT_PIDFILE = "/var/run/afancontrol.pid"
DEFAULT_INTERVAL = 5
DEFAULT_FANS_SPEED_CHECK_INTERVAL = 3
DEFAULT_HDDTEMP = "hddtemp"
DEFAULT_REPORT_CMD = (
    'printf "Subject: %s\nTo: %s\n\n%b"'
    ' "afancontrol daemon report: %REASON%" root "%MESSAGE%"'
    " | sendmail -t"
)

DEFAULT_FAN_TYPE = "linux"
DEFAULT_PWM_LINE_START = 100
DEFAULT_PWM_LINE_END = 240

DEFAULT_NEVER_STOP = True

TempName = NewType("TempName", str)
FanName = NewType("FanName", str)
MappingName = NewType("MappingName", str)

T = TypeVar("T")


FanSpeedModifier = NamedTuple(
    "FanSpeedModifier",
    # fmt: off
    [
        ("fan", FanName),
        ("modifier", float),  # [0..1]
    ]
    # fmt: on
)

FansTempsRelation = NamedTuple(
    "FansTempsRelation",
    # fmt: off
    [
        ("temps", Sequence[TempName]),
        ("fans", Sequence[FanSpeedModifier]),
    ]
    # fmt: on
)

AlertCommands = NamedTuple(
    "AlertCommands",
    # fmt: off
    [
        ("enter_cmd", Optional[str]),
        ("leave_cmd", Optional[str]),
    ]
    # fmt: on
)


Actions = NamedTuple(
    "Actions",
    # fmt: off
    [
        ("panic", AlertCommands),
        ("threshold", AlertCommands),
    ]
    # fmt: on
)


TriggerConfig = NamedTuple(
    "TriggerConfig",
    # fmt: off
    [
        ("global_commands", Actions),
        ("temp_commands", Mapping[TempName, Actions]),
    ]
    # fmt: on
)


DaemonCLIConfig = NamedTuple(
    "DaemonCLIConfig",
    # fmt: off
    [
        ("pidfile", Optional[str]),
        ("logfile", Optional[str]),
        ("exporter_listen_host", Optional[str]),
    ]
    # fmt: on
)

DaemonConfig = NamedTuple(
    "DaemonConfig",
    # fmt: off
    [
        ("pidfile", Optional[str]),
        ("logfile", Optional[str]),
        ("interval", int),
        ("exporter_listen_host", Optional[str]),
    ]
    # fmt: on
)

ParsedConfig = NamedTuple(
    "ParsedConfig",
    # fmt: off
    [
        ("daemon", DaemonConfig),
        ("report_cmd", str),
        ("triggers", TriggerConfig),
        ("fans", Mapping[FanName, PWMFanNorm]),
        ("temps", Mapping[TempName, Temp]),
        ("mappings", Mapping[MappingName, FansTempsRelation]),
    ]
    # fmt: on
)


def parse_config(config_path: Path, daemon_cli_config: DaemonCLIConfig) -> ParsedConfig:
    config = configparser.ConfigParser(interpolation=None)
    try:
        config.read_string(config_path.read_text(), source=str(config_path))
    except Exception as e:
        raise RuntimeError("Unable to parse %s:\n%s" % (config_path, e))

    daemon, hddtemp = _parse_daemon(config, daemon_cli_config)
    report_cmd, global_commands = _parse_actions(config)
    arduino_connections = _parse_arduino_connections(config)
    temps, temp_commands = _parse_temps(config, hddtemp)
    fans = _parse_fans(config, arduino_connections)
    mappings = _parse_mappings(config, fans, temps)

    return ParsedConfig(
        daemon=daemon,
        report_cmd=report_cmd,
        triggers=TriggerConfig(
            global_commands=global_commands, temp_commands=temp_commands
        ),
        fans=fans,
        temps=temps,
        mappings=mappings,
    )


def first_not_none(*parts: Optional[T]) -> Optional[T]:
    for part in parts:
        if part is not None:
            return part
    return parts[-1]  # None


def _parse_daemon(
    config: configparser.ConfigParser, daemon_cli_config: DaemonCLIConfig
) -> Tuple[DaemonConfig, str]:
    daemon = config["daemon"]
    keys = set(daemon.keys())

    pidfile = first_not_none(
        daemon_cli_config.pidfile, daemon.get("pidfile"), DEFAULT_PIDFILE
    )
    if pidfile is not None and not pidfile.strip():
        pidfile = None
    keys.discard("pidfile")

    logfile = first_not_none(daemon_cli_config.logfile, daemon.get("logfile"))
    keys.discard("logfile")

    interval = daemon.getint("interval", fallback=DEFAULT_INTERVAL)
    keys.discard("interval")

    exporter_listen_host = first_not_none(
        daemon_cli_config.exporter_listen_host, daemon.get("exporter_listen_host")
    )
    keys.discard("exporter_listen_host")

    hddtemp = daemon.get("hddtemp") or DEFAULT_HDDTEMP
    keys.discard("hddtemp")

    if keys:
        raise RuntimeError("Unknown options in the [daemon] section: %s" % (keys,))

    return (
        DaemonConfig(
            pidfile=pidfile,
            logfile=logfile,
            interval=interval,
            exporter_listen_host=exporter_listen_host,
        ),
        hddtemp,
    )


def _parse_actions(config: configparser.ConfigParser) -> Tuple[str, Actions]:
    actions = config["actions"]
    keys = set(actions.keys())

    report_cmd = first_not_none(actions.get("report_cmd"), DEFAULT_REPORT_CMD)
    assert report_cmd is not None
    keys.discard("report_cmd")

    panic = AlertCommands(
        enter_cmd=first_not_none(actions.get("panic_enter_cmd")),
        leave_cmd=first_not_none(actions.get("panic_leave_cmd")),
    )
    keys.discard("panic_enter_cmd")
    keys.discard("panic_leave_cmd")

    threshold = AlertCommands(
        enter_cmd=first_not_none(actions.get("threshold_enter_cmd")),
        leave_cmd=first_not_none(actions.get("threshold_leave_cmd")),
    )
    keys.discard("threshold_enter_cmd")
    keys.discard("threshold_leave_cmd")

    if keys:
        raise RuntimeError("Unknown options in the [actions] section: %s" % (keys,))

    return report_cmd, Actions(panic=panic, threshold=threshold)


def _parse_arduino_connections(
    config: configparser.ConfigParser
) -> Mapping[ArduinoName, ArduinoConnection]:
    arduino_connections = {}  # type: Dict[ArduinoName, ArduinoConnection]
    for section_name in config.sections():
        section_name_parts = section_name.split(":", 1)

        if section_name_parts[0].strip().lower() != "arduino":
            continue

        arduino_name = ArduinoName(section_name_parts[1].strip())
        arduino = config[section_name]
        keys = set(arduino.keys())

        serial_url = arduino["serial_url"]
        keys.discard("serial_url")

        baudrate = arduino.getint("baudrate", fallback=DEFAULT_BAUDRATE)
        keys.discard("baudrate")
        status_ttl = arduino.getint("status_ttl", fallback=DEFAULT_STATUS_TTL)
        keys.discard("status_ttl")

        if keys:
            raise RuntimeError(
                "Unknown options in the [%s] section: %s" % (section_name, keys)
            )

        if arduino_name in arduino_connections:
            raise RuntimeError(
                "Duplicate arduino section declaration for '%s'" % arduino_name
            )
        arduino_connections[arduino_name] = ArduinoConnection(
            name=arduino_name,
            serial_url=serial_url,
            baudrate=baudrate,
            status_ttl=status_ttl,
        )

    # Empty arduino_connections is ok
    return arduino_connections


def _parse_temps(
    config: configparser.ConfigParser, hddtemp: str
) -> Tuple[Mapping[TempName, Temp], Mapping[TempName, Actions]]:
    temps = {}  # type: Dict[TempName, Temp]
    temp_commands = {}  # type: Dict[TempName, Actions]
    for section_name in config.sections():
        section_name_parts = section_name.split(":", 1)

        if section_name_parts[0].strip().lower() != "temp":
            continue

        temp_name = TempName(section_name_parts[1].strip())
        temp = config[section_name]
        keys = set(temp.keys())

        actions_panic = AlertCommands(
            enter_cmd=first_not_none(temp.get("panic_enter_cmd")),
            leave_cmd=first_not_none(temp.get("panic_leave_cmd")),
        )
        keys.discard("panic_enter_cmd")
        keys.discard("panic_leave_cmd")

        actions_threshold = AlertCommands(
            enter_cmd=first_not_none(temp.get("threshold_enter_cmd")),
            leave_cmd=first_not_none(temp.get("threshold_leave_cmd")),
        )
        keys.discard("threshold_enter_cmd")
        keys.discard("threshold_leave_cmd")

        panic = TempCelsius(temp.getfloat("panic"))
        threshold = TempCelsius(temp.getfloat("threshold"))
        min = TempCelsius(temp.getfloat("min"))
        max = TempCelsius(temp.getfloat("max"))
        keys.discard("panic")
        keys.discard("threshold")
        keys.discard("min")
        keys.discard("max")

        type = temp["type"]
        keys.discard("type")

        if type == "file":
            t = FileTemp(
                temp["path"], min=min, max=max, panic=panic, threshold=threshold
            )  # type: Temp
            keys.discard("path")
        elif type == "hdd":
            if min is None or max is None:
                raise RuntimeError(
                    "hdd temp '%s' doesn't define the mandatory `min` and `max` temps"
                    % temp_name
                )
            t = HDDTemp(
                temp["path"],
                min=min,
                max=max,
                panic=panic,
                threshold=threshold,
                hddtemp_bin=hddtemp,
            )
            keys.discard("path")
        elif type == "exec":
            t = CommandTemp(
                temp["command"], min=min, max=max, panic=panic, threshold=threshold
            )
            keys.discard("command")
        else:
            raise RuntimeError(
                "Unsupported temp type '%s' for temp '%s'" % (type, temp_name)
            )

        if keys:
            raise RuntimeError(
                "Unknown options in the [%s] section: %s" % (section_name, keys)
            )

        if temp_name in temps:
            raise RuntimeError(
                "Duplicate temp section declaration for '%s'" % temp_name
            )
        temps[temp_name] = t
        temp_commands[temp_name] = Actions(
            panic=actions_panic, threshold=actions_threshold
        )

    if not temps:
        raise RuntimeError("No temps found in the config, at least 1 must be specified")
    return temps, temp_commands


def _parse_fans(
    config: configparser.ConfigParser,
    arduino_connections: Mapping[ArduinoName, ArduinoConnection],
) -> Mapping[FanName, PWMFanNorm]:
    fans = {}  # type: Dict[FanName, PWMFanNorm]
    for section_name in config.sections():
        section_name_parts = section_name.split(":", 1)

        if section_name_parts[0].strip().lower() != "fan":
            continue

        fan_name = FanName(section_name_parts[1].strip())
        fan = config[section_name]
        keys = set(fan.keys())

        fan_type = fan.get("type", fallback=DEFAULT_FAN_TYPE)
        keys.discard("type")

        if fan_type == "linux":
            pwm = PWMDevice(fan["pwm"])
            fan_input = FanInputDevice(fan["fan_input"])
            keys.discard("pwm")
            keys.discard("fan_input")

            pwmfan = LinuxPWMFan(pwm=pwm, fan_input=fan_input)  # type: BasePWMFan
        elif fan_type == "arduino":
            arduino_name = ArduinoName(fan["arduino_name"])
            keys.discard("arduino_name")
            pwm_pin = ArduinoPin(fan.getint("pwm_pin"))
            keys.discard("pwm_pin")
            tacho_pin = ArduinoPin(fan.getint("tacho_pin"))
            keys.discard("tacho_pin")

            if arduino_name not in arduino_connections:
                raise ValueError("[arduino:%s] section is missing" % arduino_name)

            pwmfan = ArduinoPWMFan(
                arduino_connections[arduino_name], pwm_pin=pwm_pin, tacho_pin=tacho_pin
            )
        else:
            raise ValueError(
                "Unsupported FAN type %s. Supported ones are "
                "`linux` and `arduino`." % fan_type
            )

        never_stop = fan.getboolean("never_stop", fallback=DEFAULT_NEVER_STOP)
        keys.discard("never_stop")

        pwm_line_start = PWMValue(
            fan.getint("pwm_line_start", fallback=DEFAULT_PWM_LINE_START)
        )
        keys.discard("pwm_line_start")

        pwm_line_end = PWMValue(
            fan.getint("pwm_line_end", fallback=DEFAULT_PWM_LINE_END)
        )
        keys.discard("pwm_line_end")

        for pwm_value in (pwm_line_start, pwm_line_end):
            if not (pwmfan.min_pwm <= pwm_value <= pwmfan.max_pwm):
                raise RuntimeError(
                    "Incorrect PWM value '%s' for fan '%s': it must be within [%s;%s]"
                    % (pwm_value, fan_name, pwmfan.min_pwm, pwmfan.max_pwm)
                )
        if pwm_line_start >= pwm_line_end:
            raise RuntimeError(
                "`pwm_line_start` PWM value must be less than `pwm_line_end` for fan '%s'"
                % (fan_name,)
            )

        if keys:
            raise RuntimeError(
                "Unknown options in the [%s] section: %s" % (section_name, keys)
            )

        if fan_name in fans:
            raise RuntimeError("Duplicate fan section declaration for '%s'" % fan_name)
        fans[fan_name] = PWMFanNorm(
            pwmfan,
            pwm_line_start=pwm_line_start,
            pwm_line_end=pwm_line_end,
            never_stop=never_stop,
        )

    if not fans:
        raise RuntimeError("No fans found in the config, at least 1 must be specified")
    return fans


def _parse_mappings(
    config: configparser.ConfigParser,
    fans: Mapping[FanName, PWMFanNorm],
    temps: Mapping[TempName, Temp],
) -> Mapping[MappingName, FansTempsRelation]:

    mappings = {}  # type: Dict[MappingName, FansTempsRelation]
    for section_name in config.sections():
        section_name_parts = section_name.split(":", 1)

        if section_name_parts[0].lower() != "mapping":
            continue

        mapping_name = MappingName(section_name_parts[1])
        mapping = config[section_name]
        keys = set(mapping.keys())

        # temps:

        mapping_temps = [
            TempName(temp_name.strip()) for temp_name in mapping["temps"].split(",")
        ]
        mapping_temps = [s for s in mapping_temps if s]
        keys.discard("temps")
        if not mapping_temps:
            raise RuntimeError(
                "Temps must not be empty in the '%s' mapping" % mapping_name
            )
        for temp_name in mapping_temps:
            if temp_name not in temps:
                raise RuntimeError(
                    "Unknown temp '%s' in mapping '%s'" % (temp_name, mapping_name)
                )
        if len(mapping_temps) != len(set(mapping_temps)):
            raise RuntimeError(
                "There are duplicate temps in mapping '%s'" % mapping_name
            )

        # fans:

        fans_with_speed = [
            fan_with_speed.strip() for fan_with_speed in mapping["fans"].split(",")
        ]
        fans_with_speed = [s for s in fans_with_speed if s]
        keys.discard("fans")

        fan_speed_pairs = [
            fan_with_speed.split("*") for fan_with_speed in fans_with_speed
        ]
        for fan_speed_pair in fan_speed_pairs:
            if len(fan_speed_pair) not in (1, 2):
                raise RuntimeError(
                    "Invalid fan specification '%s' in mapping '%s'"
                    % (fan_speed_pair, mapping_name)
                )
        mapping_fans = [
            FanSpeedModifier(
                fan=FanName(fan_speed_pair[0]),
                modifier=(
                    float(fan_speed_pair[1] if len(fan_speed_pair) == 2 else 1.0)
                ),
            )
            for fan_speed_pair in fan_speed_pairs
        ]
        for fan_speed_modifier in mapping_fans:
            if fan_speed_modifier.fan not in fans:
                raise RuntimeError(
                    "Unknown fan '%s' in mapping '%s'"
                    % (fan_speed_modifier.fan, mapping_name)
                )
            if not (0 < fan_speed_modifier.modifier <= 1.0):
                raise RuntimeError(
                    "Invalid fan modifier '%s' in mapping '%s' for fan '%s': "
                    "the allowed range is (0.0;1.0]."
                    % (
                        fan_speed_modifier.modifier,
                        mapping_name,
                        fan_speed_modifier.fan,
                    )
                )
        if len(mapping_fans) != len(
            set(fan_speed_modifier.fan for fan_speed_modifier in mapping_fans)
        ):
            raise RuntimeError(
                "There are duplicate fans in mapping '%s'" % mapping_name
            )

        if keys:
            raise RuntimeError(
                "Unknown options in the [%s] section: %s" % (section_name, keys)
            )

        if mapping_name in fans:
            raise RuntimeError(
                "Duplicate mapping section declaration for '%s'" % mapping_name
            )
        mappings[mapping_name] = FansTempsRelation(
            temps=mapping_temps, fans=mapping_fans
        )

    unused_temps = set(temps.keys())
    unused_fans = set(fans.keys())
    for relation in mappings.values():
        unused_temps -= set(relation.temps)
        unused_fans -= set(
            fan_speed_modifier.fan for fan_speed_modifier in relation.fans
        )
    if unused_temps:
        raise RuntimeError(
            "The following temps are defined but not used in any mapping: %s"
            % unused_temps
        )
    if unused_fans:
        raise RuntimeError(
            "The following fans are defined but not used in any mapping: %s"
            % unused_fans
        )
    return mappings



import logging
import os
import signal
import sys
import threading
from contextlib import ExitStack
from pathlib import Path
from typing import Optional

import click

from afancontrol.config import (
    DEFAULT_CONFIG,
    DEFAULT_PIDFILE,
    DaemonCLIConfig,
    parse_config,
)
from afancontrol.logger import logger
from afancontrol.manager import Manager
from afancontrol.metrics import Metrics, NullMetrics, PrometheusMetrics
from afancontrol.report import Report


@click.command()
@click.option("-t", "--test", is_flag=True, help="Test config")
@click.option("-d", "--daemon", is_flag=True, help="Daemonize the process after start")
@click.option("-v", "--verbose", is_flag=True, help="Increase logging verbosity")
@click.option(
    "-c",
    "--config",
    help="Config path",
    default=DEFAULT_CONFIG,
    show_default=True,
    type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    "--pidfile",
    help="Pidfile path (default is %s)" % DEFAULT_PIDFILE,
    # The default is set by the `config` module.
    type=click.Path(exists=False),
)
@click.option(
    "--logfile",
    help="Logfile path (log to stdout by default)",
    type=click.Path(exists=False),
)
@click.option(
    "--exporter-listen-host",
    help="Prometheus exporter listen host, e.g. `127.0.0.1:8000` (disabled by default)",
    type=str,
)
def daemon(
    *,
    test: bool,
    daemon: bool,
    verbose: bool,
    config: str,
    pidfile: str,
    logfile: str,
    exporter_listen_host: str
):
    """The main program of afancontrol."""

    logging.basicConfig(level=logging.DEBUG if verbose else logging.INFO)

    config_path = Path(config)
    daemon_cli_config = DaemonCLIConfig(
        pidfile=pidfile, logfile=logfile, exporter_listen_host=exporter_listen_host
    )
    parsed_config = parse_config(config_path, daemon_cli_config)

    if parsed_config.daemon.exporter_listen_host:
        metrics = PrometheusMetrics(
            parsed_config.daemon.exporter_listen_host
        )  # type: Metrics
    else:
        metrics = NullMetrics()

    manager = Manager(
        fans=parsed_config.fans,
        temps=parsed_config.temps,
        mappings=parsed_config.mappings,
        report=Report(report_command=parsed_config.report_cmd),
        triggers_config=parsed_config.triggers,
        metrics=metrics,
    )

    pidfile_instance = None  # type: Optional[PidFile]
    if parsed_config.daemon.pidfile is not None:
        pidfile_instance = PidFile(parsed_config.daemon.pidfile)

    if test:
        print("Config file '%s' is good" % config_path)
        return

    if parsed_config.daemon.logfile:
        # Logging to file should not be configured when running in
        # the config test mode.
        file_handler = logging.FileHandler(parsed_config.daemon.logfile)
        file_handler.setFormatter(
            logging.Formatter("[%(asctime)s] %(levelname)s:%(name)s:%(message)s")
        )
        logging.getLogger().addHandler(file_handler)

    signals = Signals()
    signal.signal(signal.SIGTERM, signals.sigterm)
    signal.signal(signal.SIGQUIT, signals.sigterm)
    signal.signal(signal.SIGINT, signals.sigterm)
    signal.signal(signal.SIGHUP, signals.sigterm)

    with ExitStack() as stack:
        if pidfile_instance is not None:
            stack.enter_context(pidfile_instance)
            # Ensure that pidfile is writable before forking:
            pidfile_instance.save_pid(os.getpid())

        stack.enter_context(manager)

        # Make a first tick without forking. If something is wrong,
        # (e.g. bad fan/temp file paths), an exception would be raised
        # here.
        manager.tick()

        if daemon:
            daemonize(pidfile_instance)

        while not signals.wait_for_term_queued(parsed_config.daemon.interval):
            manager.tick()


def daemonize(pidfile: Optional["PidFile"]) -> None:  # pragma: no cover
    child_pid = os.fork()
    # child_pid == 0 -- the daemonized process, which is
    #   now responsible for the `manager`'s context manager.
    #
    # child_pid != 0 -- the master process, which MUST NOT
    #   reach the end of the `with` block -- it must be terminated
    #   with the `sys.exit` call.
    if child_pid != 0:
        try:
            if pidfile is not None:
                pidfile.save_pid(child_pid)
        except Exception:
            logger.error("Unable to save process pid=%s to %s", child_pid, pidfile)
            sys.exit(1)
        else:
            sys.exit(0)


class PidFile:
    def __init__(self, pidfile: str) -> None:
        self.pidfile = Path(pidfile)

    def __str__(self):
        return "%s" % self.pidfile

    def __enter__(self):
        self.raise_if_pidfile_exists()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.remove()
        return None

    def save_pid(self, pid: int) -> None:
        self.pidfile.write_text(str(pid))

    def remove(self) -> None:
        self.pidfile.unlink()

    def raise_if_pidfile_exists(self) -> None:
        if self.pidfile.exists():
            raise RuntimeError(
                "pidfile %s already exists. Is daemon already running? "
                "Remove this file if it's not." % self
            )


class Signals:
    def __init__(self):
        self._term_event = threading.Event()

    def sigterm(self, signum, stackframe):
        self._term_event.set()

    def wait_for_term_queued(self, seconds: float) -> bool:
        is_set = self._term_event.wait(seconds)
        if is_set:
            return True
        return False



import abc
import contextlib
import threading
from http.server import BaseHTTPRequestHandler, HTTPServer
from socketserver import ThreadingMixIn
from timeit import default_timer
from typing import TYPE_CHECKING, Mapping, Optional
from urllib.parse import parse_qs, urlparse

from afancontrol.arduino import arduino_connection_from_pwmfan_norm
from afancontrol.config import TempName
from afancontrol.fans import Fans
from afancontrol.logger import logger
from afancontrol.temp import TempStatus
from afancontrol.trigger import Triggers

if TYPE_CHECKING:
    from typing import ContextManager  # Added in 3.6

try:
    import prometheus_client as prom

    prometheus_available = True
except ImportError:
    prometheus_available = False


class Metrics(abc.ABC):
    @abc.abstractmethod
    def __enter__(self):
        pass

    @abc.abstractmethod
    def __exit__(self, exc_type, exc_value, exc_tb):
        pass

    @abc.abstractmethod
    def tick(
        self,
        temps: Mapping[TempName, Optional[TempStatus]],
        fans: Fans,
        triggers: Triggers,
    ) -> None:
        pass

    @abc.abstractmethod
    def measure_tick(self) -> "ContextManager[None]":
        pass


class NullMetrics(Metrics):
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        pass

    def tick(
        self,
        temps: Mapping[TempName, Optional[TempStatus]],
        fans: Fans,
        triggers: Triggers,
    ) -> None:
        pass

    def measure_tick(self) -> "ContextManager[None]":
        @contextlib.contextmanager
        def null_context_manager():
            yield

        return null_context_manager()


class PrometheusMetrics(Metrics):
    def __init__(self, listen_host: str) -> None:
        if not prometheus_available:
            raise RuntimeError(
                "`prometheus_client` is not installed. "
                "Run `pip install 'afancontrol[metrics]'`."
            )

        self._listen_addr, port_str = listen_host.rsplit(":", 1)
        self._listen_port = int(port_str)

        self._http_server = None  # type: Optional[HTTPServer]

        self._last_metrics_collect_clock = float("nan")

        # Create a separate registry for this instance instead of using
        # the default one (which is global and doesn't allow to instantiate
        # this class more than once due to having metrics below being
        # registered for a second time):
        self.registry = prom.CollectorRegistry(auto_describe=True)

        # Register some default prometheus_client metrics:
        prom.ProcessCollector(registry=self.registry)
        if hasattr(prom, "PlatformCollector"):
            prom.PlatformCollector(registry=self.registry)
        if hasattr(prom, "GCCollector"):
            prom.GCCollector(registry=self.registry)

        # Temps:
        self.temperature_is_failing = prom.Gauge(
            "temperature_is_failing",
            "The temperature sensor is failing (it isn't returning any data)",
            ["temp_name"],
            registry=self.registry,
        )
        self.temperature_current = prom.Gauge(
            "temperature_current",
            "The current temperature value (in Celsius) from a temperature sensor",
            ["temp_name"],
            registry=self.registry,
        )
        self.temperature_min = prom.Gauge(
            "temperature_min",
            "The min temperature value (in Celsius) for a temperature sensor",
            ["temp_name"],
            registry=self.registry,
        )
        self.temperature_max = prom.Gauge(
            "temperature_max",
            "The max temperature value (in Celsius) for a temperature sensor",
            ["temp_name"],
            registry=self.registry,
        )
        self.temperature_panic = prom.Gauge(
            "temperature_panic",
            "The panic temperature value (in Celsius) for a temperature sensor",
            ["temp_name"],
            registry=self.registry,
        )
        self.temperature_threshold = prom.Gauge(
            "temperature_threshold",
            "The threshold temperature value (in Celsius) for a temperature sensor",
            ["temp_name"],
            registry=self.registry,
        )
        self.temperature_is_panic = prom.Gauge(
            "temperature_is_panic",
            "Is panic temperature reached for a temperature sensor",
            ["temp_name"],
            registry=self.registry,
        )
        self.temperature_is_threshold = prom.Gauge(
            "temperature_is_threshold",
            "Is threshold temperature reached for a temperature sensor",
            ["temp_name"],
            registry=self.registry,
        )

        # Fans:
        self.fan_rpm = prom.Gauge(
            "fan_rpm",
            "Fan speed (in RPM) as reported by the fan",
            ["fan_name"],
            registry=self.registry,
        )
        self.fan_pwm = prom.Gauge(
            "fan_pwm",
            "Current fan's PWM value (from 0 to 255)",
            ["fan_name"],
            registry=self.registry,
        )
        self.fan_pwm_normalized = prom.Gauge(
            "fan_pwm_normalized",
            "Current fan's normalized PWM value (from 0.0 to 1.0, within "
            "the `fan_pwm_line_start` and `fan_pwm_line_end` interval)",
            ["fan_name"],
            registry=self.registry,
        )
        self.fan_pwm_line_start = prom.Gauge(
            "fan_pwm_line_start",
            "PWM value where a linear correlation with RPM starts for the fan",
            ["fan_name"],
            registry=self.registry,
        )
        self.fan_pwm_line_end = prom.Gauge(
            "fan_pwm_line_end",
            "PWM value where a linear correlation with RPM ends for the fan",
            ["fan_name"],
            registry=self.registry,
        )
        self.fan_is_stopped = prom.Gauge(
            "fan_is_stopped",
            "Is PWM fan stopped because the corresponding temperatures "
            "are already low",
            ["fan_name"],
            registry=self.registry,
        )
        self.fan_is_failing = prom.Gauge(
            "fan_is_failing",
            "Is PWM fan marked as failing (e.g. because it has jammed)",
            ["fan_name"],
            registry=self.registry,
        )

        # Arduino boards:
        self.arduino_is_connected = prom.Gauge(
            "arduino_is_connected",
            "Is Arduino board connected via Serial",
            ["arduino_name"],
            registry=self.registry,
        )
        self.arduino_status_age_seconds = prom.Gauge(
            "arduino_status_age_seconds",
            "Seconds since the last `status` message from "
            "the Arduino board (measured at the latest tick)",
            ["arduino_name"],
            registry=self.registry,
        )

        # Others:
        self.is_panic = prom.Gauge(
            "is_panic", "Is in panic mode", registry=self.registry
        )
        self.is_threshold = prom.Gauge(
            "is_threshold", "Is in threshold mode", registry=self.registry
        )

        self.tick_duration = prom.Histogram(
            # Summary would have been better there, but prometheus_client
            # doesn't yet support quantiles in Summaries.
            # See: https://github.com/prometheus/client_python/issues/92
            "tick_duration",
            "Duration of a single tick",
            buckets=(0.1, 0.25, 0.5, 0.75, 1.0, 2.5, 5.0, 10.0, float("inf")),
            registry=self.registry,
        )
        last_metrics_tick_seconds_ago = prom.Gauge(
            "last_metrics_tick_seconds_ago",
            "The time in seconds since the last tick (which also updates these metrics)",
            registry=self.registry,
        )
        last_metrics_tick_seconds_ago.set_function(
            lambda: self.last_metrics_tick_seconds_ago
        )

    @property
    def last_metrics_tick_seconds_ago(self):
        return self._clock() - self._last_metrics_collect_clock

    def _start(self):
        # `prometheus_client.start_http_server` which persists a server reference
        # so it could be stopped later.
        CustomMetricsHandler = MetricsHandler.factory(self.registry)
        httpd = _ThreadingSimpleServer(
            (self._listen_addr, self._listen_port), CustomMetricsHandler
        )
        t = threading.Thread(target=httpd.serve_forever)
        t.daemon = True
        t.start()
        return httpd

    def __enter__(self):
        self._http_server = self._start()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._http_server.shutdown()  # stop serve_forever()
        self._http_server.server_close()
        self._http_server = None
        return None

    def tick(
        self,
        temps: Mapping[TempName, Optional[TempStatus]],
        fans: Fans,
        triggers: Triggers,
    ) -> None:
        for temp_name, temp_status in temps.items():
            if temp_status is None:
                self.temperature_is_failing.labels(temp_name).set(1)
                self.temperature_current.labels(temp_name).set(none_to_nan(None))
                self.temperature_min.labels(temp_name).set(none_to_nan(None))
                self.temperature_max.labels(temp_name).set(none_to_nan(None))
                self.temperature_panic.labels(temp_name).set(none_to_nan(None))
                self.temperature_threshold.labels(temp_name).set(none_to_nan(None))
                self.temperature_is_panic.labels(temp_name).set(none_to_nan(None))
                self.temperature_is_threshold.labels(temp_name).set(none_to_nan(None))
            else:
                self.temperature_is_failing.labels(temp_name).set(0)
                self.temperature_current.labels(temp_name).set(temp_status.temp)
                self.temperature_min.labels(temp_name).set(temp_status.min)
                self.temperature_max.labels(temp_name).set(temp_status.max)
                self.temperature_panic.labels(temp_name).set(
                    none_to_nan(temp_status.panic)
                )
                self.temperature_threshold.labels(temp_name).set(
                    none_to_nan(temp_status.threshold)
                )
                self.temperature_is_panic.labels(temp_name).set(temp_status.is_panic)
                self.temperature_is_threshold.labels(temp_name).set(
                    temp_status.is_threshold
                )

        for fan_name, pwmfan_norm in fans.fans.items():
            self._collect_fan_metrics(fans, fan_name, pwmfan_norm)

        arduino_connections = {
            arduino_connection.name: arduino_connection
            for arduino_connection in (
                arduino_connection_from_pwmfan_norm(pwmfan_norm)
                for pwmfan_norm in fans.fans.values()
            )
            if arduino_connection is not None
        }
        for arduino_name, arduino_connection in arduino_connections.items():
            self.arduino_is_connected.labels(arduino_name).set(
                arduino_connection.is_connected
            )
            self.arduino_status_age_seconds.labels(arduino_name).set(
                arduino_connection.status_age_seconds
            )

        self.is_panic.set(triggers.panic_trigger.is_alerting)
        self.is_threshold.set(triggers.threshold_trigger.is_alerting)

        self._last_metrics_collect_clock = self._clock()

    def measure_tick(self) -> "ContextManager[None]":
        return self.tick_duration.time()

    def _collect_fan_metrics(self, fans, fan_name, pwm_fan_norm):
        self.fan_pwm_line_start.labels(fan_name).set(pwm_fan_norm.pwm_line_start)
        self.fan_pwm_line_end.labels(fan_name).set(pwm_fan_norm.pwm_line_end)
        self.fan_is_stopped.labels(fan_name).set(fans.is_fan_stopped(fan_name))
        self.fan_is_failing.labels(fan_name).set(fans.is_fan_failing(fan_name))
        try:
            self.fan_rpm.labels(fan_name).set(pwm_fan_norm.get_speed())
            self.fan_pwm.labels(fan_name).set(pwm_fan_norm.get_raw())
            self.fan_pwm_normalized.labels(fan_name).set(pwm_fan_norm.get())
        except Exception:
            logger.warning(
                "Failed to collect metrics for fan %s", fan_name, exc_info=True
            )
            self.fan_rpm.labels(fan_name).set(none_to_nan(None))
            self.fan_pwm.labels(fan_name).set(none_to_nan(None))
            self.fan_pwm_normalized.labels(fan_name).set(none_to_nan(None))

    def _clock(self):
        return default_timer()


def none_to_nan(v: Optional[float]) -> float:
    if v is None:
        return float("nan")
    return v


class _ThreadingSimpleServer(ThreadingMixIn, HTTPServer):
    """Thread per request HTTP server."""

    # https://github.com/prometheus/client_python/blob/31f5557e2e84ca4ffa9a03abf6e3f4d0c8b8c3eb/prometheus_client/exposition.py#L180-L187  # noqa
    #
    # Make worker threads "fire and forget". Beginning with Python 3.7 this
    # prevents a memory leak because ``ThreadingMixIn`` starts to gather all
    # non-daemon threads in a list in order to join on them at server close.
    # Enabling daemon threads virtually makes ``_ThreadingSimpleServer`` the
    # same as Python 3.7's ``ThreadingHTTPServer``.
    daemon_threads = True


# `MetricsHandler` of `prometheus_client==0.0.18` doesn't support exposing
# a custom registry. This is backported below:
if prometheus_available:
    if hasattr(prom.MetricsHandler, "factory"):
        MetricsHandler = prom.MetricsHandler
    else:
        from prometheus_client.exposition import generate_latest, CONTENT_TYPE_LATEST

        class MetricsHandler(BaseHTTPRequestHandler):  # type: ignore
            # https://github.com/prometheus/client_python/blob/31f5557e2e84ca4ffa9a03abf6e3f4d0c8b8c3eb/prometheus_client/exposition.py#L141-L177  # noqa
            registry = prom.REGISTRY

            def do_GET(self):
                registry = self.registry
                params = parse_qs(urlparse(self.path).query)
                if "name[]" in params:
                    registry = registry.restricted_registry(params["name[]"])
                try:
                    output = generate_latest(registry)
                except Exception:
                    self.send_error(500, "error generating metric output")
                    raise
                self.send_response(200)
                self.send_header("Content-Type", CONTENT_TYPE_LATEST)
                self.end_headers()
                self.wfile.write(output)

            def log_message(self, format, *args):
                """Log nothing."""

            @classmethod
            def factory(cls, registry):
                cls_name = str(cls.__name__)
                MyMetricsHandler = type(cls_name, (cls, object), {"registry": registry})
                return MyMetricsHandler



from contextlib import ExitStack

import pytest

from afancontrol.pwmfan import (
    FanInputDevice,
    LinuxPWMFan,
    PWMDevice,
    PWMFanNorm,
    PWMValue,
)


@pytest.fixture
def pwm_path(temp_path):
    # pwm = /sys/class/hwmon/hwmon0/pwm2
    pwm_path = temp_path / "pwm2"
    pwm_path.write_text("0\n")
    return pwm_path


@pytest.fixture
def pwm_enable_path(temp_path):
    pwm_enable_path = temp_path / "pwm2_enable"
    pwm_enable_path.write_text("0\n")
    return pwm_enable_path


@pytest.fixture
def fan_input_path(temp_path):
    # fan_input = /sys/class/hwmon/hwmon0/fan2_input
    fan_input_path = temp_path / "fan2_input"
    fan_input_path.write_text("1300\n")
    return fan_input_path


@pytest.fixture
def pwmfan_norm(pwmfan):
    return PWMFanNorm(
        pwmfan,
        pwm_line_start=PWMValue(100),
        pwm_line_end=PWMValue(240),
        never_stop=False,
    )


@pytest.fixture
def pwmfan(pwm_path, fan_input_path):
    return LinuxPWMFan(
        pwm=PWMDevice(str(pwm_path)), fan_input=FanInputDevice(str(fan_input_path))
    )


@pytest.mark.parametrize("pwmfan_fixture", ["pwmfan", "pwmfan_norm"])
def test_get_speed(pwmfan_fixture, pwmfan, pwmfan_norm, fan_input_path):
    fan = locals()[pwmfan_fixture]
    fan_input_path.write_text("721\n")
    assert 721 == fan.get_speed()


@pytest.mark.parametrize("pwmfan_fixture", ["pwmfan", "pwmfan_norm"])
@pytest.mark.parametrize("raises", [True, False])
def test_enter_exit(
    raises, pwmfan_fixture, pwmfan, pwmfan_norm, pwm_enable_path, pwm_path
):
    fan = locals()[pwmfan_fixture]

    class Exc(Exception):
        pass

    with ExitStack() as stack:
        if raises:
            stack.enter_context(pytest.raises(Exc))
        stack.enter_context(fan)

        assert "1" == pwm_enable_path.read_text()
        assert "255" == pwm_path.read_text()

        value = dict(pwmfan=100, pwmfan_norm=0.39)[pwmfan_fixture]  # 100/255 ~= 0.39
        fan.set(value)

        assert "1" == pwm_enable_path.read_text()
        assert "100" == pwm_path.read_text()

        if raises:
            raise Exc()

    assert "0" == pwm_enable_path.read_text()
    assert "100" == pwm_path.read_text()  # `fancontrol` doesn't reset speed


def test_get_set_pwmfan(pwmfan, pwm_path):
    pwmfan.set(142)
    assert "142" == pwm_path.read_text()

    pwm_path.write_text("132\n")
    assert 132 == pwmfan.get()

    pwmfan.set_full_speed()
    assert "255" == pwm_path.read_text()

    with pytest.raises(ValueError):
        pwmfan.set(256)

    with pytest.raises(ValueError):
        pwmfan.set(-1)


def test_get_set_pwmfan_norm(pwmfan_norm, pwm_path):
    pwmfan_norm.set(0.42)
    assert "101" == pwm_path.read_text()

    pwm_path.write_text("132\n")
    assert pytest.approx(0.517, 0.01) == pwmfan_norm.get()

    pwmfan_norm.set_full_speed()
    assert "255" == pwm_path.read_text()

    assert 238 == pwmfan_norm.set(0.99)
    assert "238" == pwm_path.read_text()

    assert 255 == pwmfan_norm.set(1.0)
    assert "255" == pwm_path.read_text()

    assert 255 == pwmfan_norm.set(1.1)
    assert "255" == pwm_path.read_text()

    assert 0 == pwmfan_norm.set(-0.1)
    assert "0" == pwm_path.read_text()



import click

import afancontrol
from afancontrol.daemon import daemon
from afancontrol.fantest import fantest


@click.group()
@click.version_option(version=afancontrol.__version__)
def main():
    """afancontrol is an Advanced Fan Control program, which controls PWM
    fans according to the current temperatures of the system components.
    """
    pass


main.add_command(daemon)
main.add_command(fantest)

if __name__ == "__main__":
    main(prog_name="afancontrol")



from unittest.mock import MagicMock, call

import pytest

from afancontrol import trigger
from afancontrol.config import Actions, AlertCommands, TempName, TriggerConfig
from afancontrol.report import Report
from afancontrol.temp import TempCelsius, TempStatus
from afancontrol.trigger import PanicTrigger, ThresholdTrigger, Triggers


@pytest.fixture
def report():
    return MagicMock(spec=Report)


def test_panic_on_empty_temp(report, sense_exec_shell_command):
    t = PanicTrigger(
        global_commands=AlertCommands(
            enter_cmd="printf '@%s' enter", leave_cmd="printf '@%s' leave"
        ),
        temp_commands={
            TempName("mobo"): AlertCommands(
                enter_cmd=None, leave_cmd="printf '@%s' mobo leave"
            )
        },
        report=report,
    )

    with sense_exec_shell_command(trigger) as (mock_exec_shell_command, get_stdout):
        with t:
            assert not t.is_alerting
            assert 0 == mock_exec_shell_command.call_count
            t.check({TempName("mobo"): None})
            assert t.is_alerting

            assert mock_exec_shell_command.call_args_list == [
                call("printf '@%s' enter")
            ]
            assert ["@enter"] == get_stdout()
            mock_exec_shell_command.reset_mock()

        assert not t.is_alerting
        assert mock_exec_shell_command.call_args_list == [
            call("printf '@%s' mobo leave"),
            call("printf '@%s' leave"),
        ]
        assert ["@mobo@leave", "@leave"] == get_stdout()


def test_threshold_on_empty_temp(report):
    t = ThresholdTrigger(
        global_commands=AlertCommands(enter_cmd=None, leave_cmd=None),
        temp_commands={TempName("mobo"): AlertCommands(enter_cmd=None, leave_cmd=None)},
        report=report,
    )
    with t:
        assert not t.is_alerting
        t.check({TempName("mobo"): None})
        assert not t.is_alerting
    assert not t.is_alerting


@pytest.mark.parametrize("cls", [ThresholdTrigger, PanicTrigger])
def test_good_temp(cls, report):
    t = cls(
        global_commands=AlertCommands(enter_cmd=None, leave_cmd=None),
        temp_commands=dict(mobo=AlertCommands(enter_cmd=None, leave_cmd=None)),
        report=report,
    )
    with t:
        assert not t.is_alerting
        t.check(
            dict(
                mobo=TempStatus(
                    temp=TempCelsius(34.0),
                    min=TempCelsius(40.0),
                    max=TempCelsius(50.0),
                    panic=TempCelsius(60.0),
                    threshold=None,
                    is_panic=False,
                    is_threshold=False,
                )
            )
        )
        assert not t.is_alerting


@pytest.mark.parametrize("cls", [ThresholdTrigger, PanicTrigger])
def test_bad_temp(cls, report, sense_exec_shell_command):
    t = cls(
        global_commands=AlertCommands(
            enter_cmd="printf '@%s' enter", leave_cmd="printf '@%s' leave"
        ),
        temp_commands=dict(
            mobo=AlertCommands(
                enter_cmd="printf '@%s' mobo enter", leave_cmd="printf '@%s' mobo leave"
            )
        ),
        report=report,
    )
    with sense_exec_shell_command(trigger) as (mock_exec_shell_command, get_stdout):
        with t:
            assert not t.is_alerting
            t.check(
                dict(
                    mobo=TempStatus(
                        temp=TempCelsius(70.0),
                        min=TempCelsius(40.0),
                        max=TempCelsius(50.0),
                        panic=TempCelsius(60.0),
                        threshold=TempCelsius(55.0),
                        is_panic=True,
                        is_threshold=True,
                    )
                )
            )
            assert t.is_alerting
            assert mock_exec_shell_command.call_args_list == [
                call("printf '@%s' mobo enter"),
                call("printf '@%s' enter"),
            ]
            assert ["@mobo@enter", "@enter"] == get_stdout()
            mock_exec_shell_command.reset_mock()

            t.check(
                dict(
                    mobo=TempStatus(
                        temp=TempCelsius(34.0),
                        min=TempCelsius(40.0),
                        max=TempCelsius(50.0),
                        panic=TempCelsius(60.0),
                        threshold=None,
                        is_panic=False,
                        is_threshold=False,
                    )
                )
            )
            assert not t.is_alerting
            assert mock_exec_shell_command.call_args_list == [
                call("printf '@%s' mobo leave"),
                call("printf '@%s' leave"),
            ]
            assert ["@mobo@leave", "@leave"] == get_stdout()
            mock_exec_shell_command.reset_mock()
        assert 0 == mock_exec_shell_command.call_count


def test_triggers_good_temp(report):
    t = Triggers(
        TriggerConfig(
            global_commands=Actions(
                panic=AlertCommands(enter_cmd=None, leave_cmd=None),
                threshold=AlertCommands(enter_cmd=None, leave_cmd=None),
            ),
            temp_commands={
                TempName("mobo"): Actions(
                    panic=AlertCommands(enter_cmd=None, leave_cmd=None),
                    threshold=AlertCommands(enter_cmd=None, leave_cmd=None),
                )
            },
        ),
        report=report,
    )
    with t:
        assert not t.is_alerting
        t.check(
            {
                TempName("mobo"): TempStatus(
                    temp=TempCelsius(34.0),
                    min=TempCelsius(40.0),
                    max=TempCelsius(50.0),
                    panic=TempCelsius(60.0),
                    threshold=None,
                    is_panic=False,
                    is_threshold=False,
                )
            }
        )
        assert not t.is_alerting



import subprocess

import pytest

from afancontrol.exec import exec_shell_command


def test_exec_shell_command_successful():
    assert "42\n" == exec_shell_command("echo 42")


def test_exec_shell_command_ignores_stderr():
    assert "42\n" == exec_shell_command("echo 111 >&2; echo 42")


def test_exec_shell_command_erroneous():
    with pytest.raises(subprocess.SubprocessError):
        exec_shell_command("echo 42 && false")


def test_exec_shell_command_raises_for_unicode():
    with pytest.raises(ValueError):
        exec_shell_command("echo привет")


def test_exec_shell_command_expands_glob(temp_path):
    (temp_path / "sda").write_text("")
    (temp_path / "sdb").write_text("")

    expected = "{0}/sda {0}/sdb\n".format(temp_path)
    assert expected == exec_shell_command('echo "%s/sd"?' % temp_path)



import json
import queue
import struct
import threading
from timeit import default_timer
from typing import Any, Dict, NewType, Optional

from afancontrol.logger import logger
from afancontrol.pwmfan import BasePWMFan, FanValue, PWMFanNorm, PWMValue

try:
    from serial import serial_for_url
    from serial.threaded import LineReader, ReaderThread

    pyserial_available = True
except ImportError:
    LineReader = object
    ReaderThread = object

    pyserial_available = False

ArduinoName = NewType("ArduinoName", str)
ArduinoPin = NewType("ArduinoPin", int)

DEFAULT_BAUDRATE = 115200
DEFAULT_STATUS_TTL = 5


def arduino_connection_from_pwmfan_norm(
    pwmfan_norm: PWMFanNorm
) -> Optional["ArduinoConnection"]:
    # Used in metrics
    pwmfan = pwmfan_norm.pwmfan
    if isinstance(pwmfan, ArduinoPWMFan):
        return pwmfan._conn
    return None


class ArduinoPWMFan(BasePWMFan):
    def __init__(
        self,
        arduino_connection: "ArduinoConnection",
        *,
        pwm_pin: ArduinoPin,
        tacho_pin: ArduinoPin
    ) -> None:
        super().__init__()
        self._conn = arduino_connection
        self._pwm_pin = pwm_pin
        self._tacho_pin = tacho_pin

    def get(self) -> PWMValue:
        return PWMValue(int(self._conn.get_pwm(self._pwm_pin)))

    def _set_raw(self, pwm: PWMValue) -> None:
        self._conn.set_pwm(self._pwm_pin, pwm)

    def get_speed(self) -> FanValue:
        return FanValue(self._conn.get_rpm(self._tacho_pin))

    def __enter__(self):  # reusable
        self._conn.__enter__()
        super().__enter__()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        try:
            super().__exit__(exc_type, exc_value, exc_tb)
        finally:
            self._conn.__exit__(exc_type, exc_value, exc_tb)

    def _enable_pwm(self) -> None:
        self.set_full_speed()

    def _disable_pwm(self) -> None:
        self.set_full_speed()
        self._conn.wait_for_status()

        if self.get() >= type(self).max_pwm:
            return

        raise RuntimeError("Couldn't disable PWM on the fan %r" % self)

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (
                self._conn == other._conn
                and self._pwm_pin == other._pwm_pin
                and self._tacho_pin == other._tacho_pin
            )

        return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s(%r, pwm_pin=%r, tacho_pin=%r)" % (
            type(self).__name__,
            self._conn,
            self._pwm_pin,
            self._tacho_pin,
        )


class ArduinoConnection:
    def __init__(
        self,
        name: ArduinoName,
        serial_url: str,
        *,
        baudrate: int = DEFAULT_BAUDRATE,
        status_ttl: int = DEFAULT_STATUS_TTL
    ) -> None:
        if not pyserial_available:
            raise RuntimeError(
                "`pyserial` is not installed. "
                "Run `pip install 'afancontrol[arduino]'`."
            )
        self.name = name
        self.url = serial_url
        self.baudrate = baudrate
        self.status_ttl = status_ttl
        self._reader_thread = _AutoRetriedReaderThread(
            lambda: _StatusProtocol(self), url=serial_url, baudrate=baudrate
        )
        self._context_manager_depth = 0
        self._status = None  # type: Optional[Dict[str, Dict[str, int]]]
        self._status_clock = None  # type: Optional[float]
        self._status_lock = threading.Lock()
        self._status_event = threading.Event()

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (
                self.name == other.name
                and self.url == other.url
                and self.baudrate == other.baudrate
                and self.status_ttl == other.status_ttl
            )

        return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s(%r, %r, baudrate=%r, status_ttl=%r)" % (
            type(self).__name__,
            self.name,
            self.url,
            self.baudrate,
            self.status_ttl,
        )

    def __enter__(self):  # reentrant
        if self._context_manager_depth == 0:
            self._reader_thread.__enter__()
        self._context_manager_depth += 1
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._context_manager_depth -= 1
        if self._context_manager_depth == 0:
            return self._reader_thread.__exit__(exc_type, exc_value, exc_tb)
        return None

    def _clock(self):
        return default_timer()

    def _incoming_message(self, message: Dict[str, Any]) -> None:
        # Called by the pyserial Protocol `_StatusProtocol`.
        if "error" in message:
            logger.warning("Received an error from Arduino %s: %r", self.url, message)
        else:
            self._update_status(message)

    def _update_status(self, status: Dict[str, Dict[str, int]]) -> None:
        with self._status_lock:
            self._status = status
            self._status_clock = self._clock()
        self._status_event.set()

    @property
    def is_connected(self) -> bool:
        try:
            with self._status_lock:
                self._ensure_status_is_valid()
        except Exception:
            return False
        else:
            return True

    def get_rpm(self, pin: ArduinoPin) -> int:
        if self._status is None:
            self.wait_for_status()
        with self._status_lock:
            self._ensure_status_is_valid()
            assert self._status is not None
            return int(self._status["fan_inputs"][str(pin)])

    def get_pwm(self, pin: ArduinoPin) -> int:
        if self._status is None:
            self.wait_for_status()
        with self._status_lock:
            self._ensure_status_is_valid()
            assert self._status is not None
            return int(self._status["fan_pwm"][str(pin)])

    def _ensure_status_is_valid(self):
        if self._status is None:
            raise RuntimeError("No status from the Arduino board at %s" % self.url)
        assert self._status_clock is not None
        status_age = self._clock() - self._status_clock
        if status_age > self.status_ttl:
            self._reader_thread.check_connection()
            raise RuntimeError(
                "The last received status from the Arduino board "
                "at %s was too long ago: %s seconds" % (self.url, status_age)
            )

    @property
    def status_age_seconds(self) -> float:
        with self._status_lock:
            if self._status_clock is None:
                return float("nan")
            return self._clock() - self._status_clock

    def set_pwm(self, pin: ArduinoPin, pwm: PWMValue) -> None:
        command = SetPWMCommand(pwm_pin=pin, pwm=pwm).to_bytes()
        transport = self._reader_thread.transport
        try:
            transport.write(command)
            transport.flush()
        except Exception:
            self._reader_thread.check_connection()
            raise

    def wait_for_status(self) -> None:
        self._status_event.clear()
        if self._status_event.wait(self.status_ttl) is not True:
            raise RuntimeError(
                "Timed out waiting for the status from Arduino board at %s" % self.url
            )


class SetPWMCommand:
    command = b"\xf1"

    def __init__(self, *, pwm_pin: ArduinoPin, pwm: PWMValue) -> None:
        self.pwm_pin = pwm_pin
        self.pwm = pwm

    def __repr__(self):
        return "%s(pwm_pin=%r, pwm=%r)" % (type(self).__name__, self.pwm_pin, self.pwm)

    def to_bytes(self):
        return struct.pack("sBB", self.command, self.pwm_pin, self.pwm)

    @classmethod
    def parse(cls, b: bytes) -> "SetPWMCommand":
        command, pwm_pin, pwm = struct.unpack("sBB", b)
        if command != cls.command:
            raise ValueError(
                "Invalid command marker. Expected %r, got %r" % (cls.command, command)
            )
        return cls(pwm_pin=ArduinoPin(pwm_pin), pwm=PWMValue(pwm))


class _StatusProtocol(LineReader):
    TERMINATOR = b"\n"

    def __init__(self, arduino_connection: ArduinoConnection) -> None:
        super().__init__()
        self._arduino_connection = arduino_connection

    def handle_line(self, line: str) -> None:
        try:
            message = json.loads(line)
            self._arduino_connection._incoming_message(message)
        except Exception:  # `handle_line` should not raise exceptions
            logger.error(
                "Unable to parse the status line from Arduino as json: %r",
                line,
                exc_info=True,
            )


class _AutoRetriedReaderThread:
    _QUEUE_STOP = object()
    _QUEUE_CHECK = object()

    def __init__(self, protocol_factory, **serial_for_url_kwargs) -> None:
        self.protocol_factory = protocol_factory
        self.serial_for_url_kwargs = serial_for_url_kwargs
        self._reader_thread = None  # type: Optional[ReaderThread]
        self._transport = None  # type: Optional[ReaderThread]
        self._watchdog_thread = None  # type: Optional[threading.Thread]
        self._watchdog_queue = queue.Queue()  # type: queue.Queue[Any]

    def __enter__(self):  # reusable
        # TODO ?? maybe clean the _watchdog_queue?
        self._reader_thread, self._transport = self._new_reader_thread()
        self._watchdog_thread = threading.Thread(target=self._thread_run, daemon=True)
        self._watchdog_thread.start()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._watchdog_queue.put(self._QUEUE_STOP)
        self._watchdog_thread.join()
        self._reader_thread.close()
        self._reader_thread = None
        self._transport = None

    @property
    def transport(self):
        return self._transport

    def check_connection(self):
        self._watchdog_queue.put(self._QUEUE_CHECK)

    def _new_reader_thread(self):
        ser = serial_for_url(**self.serial_for_url_kwargs)
        thread = _ReaderThreadWithFlush(ser, self.protocol_factory)
        thread.start()
        transport, _ = thread.connect()
        return thread, transport

    def _thread_run(self):
        while True:
            item = self._watchdog_queue.get()
            try:
                if self._reader_thread is None:
                    break
                if item is self._QUEUE_STOP:
                    break
                elif item is self._QUEUE_CHECK:
                    if self._reader_thread.alive:
                        continue
                    self._reader_thread.close()
                    self._reader_thread, self._transport = self._new_reader_thread()
            except Exception:  # `_thread_run` should not raise
                logger.error(
                    "Error in the Arduino connection watchdog thread", exc_info=True
                )
            finally:
                self._watchdog_queue.task_done()


class _ReaderThreadWithFlush(ReaderThread):
    def flush(self):
        with self._lock:
            self.serial.flush()



import subprocess

from afancontrol.logger import logger


def exec_shell_command(shell_command: str, timeout: int = 5) -> str:
    try:
        p = subprocess.run(
            shell_command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
            check=True,
            timeout=timeout,
        )
        out = p.stdout.decode("ascii")
        err = p.stderr.decode().strip()
        if err:
            logger.warning(
                "Shell command '%s' executed successfully, but printed to stderr:\n%s",
                shell_command,
                err,
            )
        return out
    except subprocess.CalledProcessError as e:
        ec = e.returncode
        out = e.stdout.decode().strip()
        err = e.stderr.decode().strip()
        logger.error(
            "Shell command '%s' failed (exit code %s):\nstdout:\n%s\nstderr:\n%s\n",
            shell_command,
            ec,
            out,
            err,
        )
        raise



import random
import types
from time import sleep
from unittest.mock import MagicMock

import pytest
import requests

from afancontrol.config import FanName, TempName
from afancontrol.fans import Fans
from afancontrol.metrics import PrometheusMetrics, prometheus_available
from afancontrol.pwmfan import PWMFanNorm
from afancontrol.report import Report
from afancontrol.temp import TempCelsius, TempStatus
from afancontrol.trigger import Triggers


@pytest.fixture
def requests_session():
    # Ignore system proxies, see https://stackoverflow.com/a/28521696
    with requests.Session() as session:
        session.trust_env = False
        yield session


@pytest.mark.skipif(
    not prometheus_available, reason="prometheus_client is not installed"
)
def test_prometheus_metrics(requests_session):
    mocked_fan = MagicMock(spec=PWMFanNorm)()
    mocked_triggers = MagicMock(spec=Triggers)()
    mocked_report = MagicMock(spec=Report)()

    port = random.randint(20000, 50000)
    metrics = PrometheusMetrics("127.0.0.1:%s" % port)
    with metrics:
        resp = requests_session.get("http://127.0.0.1:%s/metrics" % port)
        assert resp.status_code == 200
        assert "is_threshold 0.0" in resp.text

        with metrics.measure_tick():
            sleep(0.01)

        resp = requests_session.get("http://127.0.0.1:%s/metrics" % port)
        assert resp.status_code == 200
        assert "tick_duration_count 1.0" in resp.text
        assert "tick_duration_sum 0." in resp.text

        mocked_triggers.panic_trigger.is_alerting = True
        mocked_triggers.threshold_trigger.is_alerting = False

        mocked_fan.pwm_line_start = 100
        mocked_fan.pwm_line_end = 240
        mocked_fan.get_speed.return_value = 999
        mocked_fan.get_raw.return_value = 142
        mocked_fan.get = types.MethodType(PWMFanNorm.get, mocked_fan)
        mocked_fan.pwmfan.max_pwm = 255

        metrics.tick(
            temps={
                TempName("goodtemp"): TempStatus(
                    temp=TempCelsius(74.0),
                    min=TempCelsius(40.0),
                    max=TempCelsius(50.0),
                    panic=TempCelsius(60.0),
                    threshold=None,
                    is_panic=True,
                    is_threshold=False,
                ),
                TempName("failingtemp"): None,
            },
            fans=Fans(fans={FanName("test"): mocked_fan}, report=mocked_report),
            triggers=mocked_triggers,
        )

        resp = requests_session.get("http://127.0.0.1:%s/metrics" % port)
        assert resp.status_code == 200
        print(resp.text)
        assert 'temperature_current{temp_name="failingtemp"} NaN' in resp.text
        assert 'temperature_current{temp_name="goodtemp"} 74.0' in resp.text
        assert 'temperature_is_failing{temp_name="failingtemp"} 1.0' in resp.text
        assert 'temperature_is_failing{temp_name="goodtemp"} 0.0' in resp.text
        assert 'fan_rpm{fan_name="test"} 999.0' in resp.text
        assert 'fan_pwm{fan_name="test"} 142.0' in resp.text
        assert 'fan_pwm_normalized{fan_name="test"} 0.556' in resp.text
        assert 'fan_is_failing{fan_name="test"} 0.0' in resp.text
        assert "is_panic 1.0" in resp.text
        assert "is_threshold 0.0" in resp.text
        assert "last_metrics_tick_seconds_ago 0." in resp.text

    with pytest.raises(IOError):
        requests_session.get("http://127.0.0.1:%s/metrics" % port)


@pytest.mark.skipif(
    not prometheus_available, reason="prometheus_client is not installed"
)
def test_prometheus_faulty_fans_dont_break_metrics_collection(requests_session):
    mocked_fan = MagicMock(spec=PWMFanNorm)()
    mocked_triggers = MagicMock(spec=Triggers)()
    mocked_report = MagicMock(spec=Report)()

    port = random.randint(20000, 50000)
    metrics = PrometheusMetrics("127.0.0.1:%s" % port)
    with metrics:
        mocked_triggers.panic_trigger.is_alerting = False
        mocked_triggers.threshold_trigger.is_alerting = False

        mocked_fan.pwm_line_start = 100
        mocked_fan.pwm_line_end = 240
        mocked_fan.get_speed.side_effect = IOError
        mocked_fan.get_raw.side_effect = IOError

        # Must not raise despite the PWMFan methods raising above:
        metrics.tick(
            temps={TempName("failingtemp"): None},
            fans=Fans(fans={FanName("test"): mocked_fan}, report=mocked_report),
            triggers=mocked_triggers,
        )

        resp = requests_session.get("http://127.0.0.1:%s/metrics" % port)
        assert resp.status_code == 200
        assert 'fan_pwm_line_start{fan_name="test"} 100.0' in resp.text
        assert 'fan_pwm_line_end{fan_name="test"} 240.0' in resp.text
        assert 'fan_rpm{fan_name="test"} NaN' in resp.text
        assert 'fan_pwm{fan_name="test"} NaN' in resp.text
        assert 'fan_pwm_normalized{fan_name="test"} NaN' in resp.text
        assert 'fan_is_failing{fan_name="test"} 0.0' in resp.text
        assert "is_panic 0.0" in resp.text
        assert "is_threshold 0.0" in resp.text



import subprocess
from typing import Optional
from unittest.mock import patch

import pytest

from afancontrol.temp import (
    CommandTemp,
    FileTemp,
    HDDTemp,
    Temp,
    TempCelsius,
    TempStatus,
)


@pytest.fixture
def file_temp_path(temp_path):
    # /sys/class/hwmon/hwmon0/temp1_input
    temp_input_path = temp_path / "temp1_input"
    temp_input_path.write_text("34000\n")

    temp_max_path = temp_path / "temp1_max"
    temp_max_path.write_text("127000\n")

    temp_min_path = temp_path / "temp1_min"
    # My mobo actually returns this as min:
    temp_min_path.write_text("127000\n")

    return temp_input_path


@pytest.fixture
def hddtemp_output_many():
    return (
        "/dev/sda: Adaptec XXXXX:  drive supported,"
        " but it doesn't have a temperature sensor.\n"
        "/dev/sdb: Adaptec XXXXX:  drive supported,"
        " but it doesn't have a temperature sensor.\n"
        "38\n"
        "39\n"
        "30\n"
        "36\n"
    )


@pytest.fixture
def hddtemp_output_bad():
    return (
        "/dev/sda: Adaptec XXXXX:  drive supported,"
        " but it doesn't have a temperature sensor.\n"
    )


class DummyTemp(Temp):
    def _get_temp(self):
        pass


@pytest.mark.parametrize(
    "temp, threshold, panic, is_threshold, is_panic",
    [
        (34.0, None, 60.0, False, False),
        (42.0, None, 60.0, False, False),
        (57.0, 55.0, 60.0, True, False),
        (61.0, 55.0, 61.0, True, True),
        (61.0, None, 61.0, False, True),
    ],
)
def test_temp(
    temp: TempCelsius,
    threshold: Optional[TempCelsius],
    panic: TempCelsius,
    is_threshold,
    is_panic,
):
    min = TempCelsius(40.0)
    max = TempCelsius(50.0)

    with patch.object(DummyTemp, "_get_temp") as mock_get_temp:
        t = DummyTemp(panic=panic, threshold=threshold)
        mock_get_temp.return_value = [temp, min, max]

        assert t.get() == TempStatus(
            temp=temp,
            min=min,
            max=max,
            panic=panic,
            threshold=threshold,
            is_panic=is_panic,
            is_threshold=is_threshold,
        )


def test_file_temp_min_max_numbers(file_temp_path):
    temp = FileTemp(
        temp_path=str(file_temp_path),
        min=TempCelsius(40.0),
        max=TempCelsius(50.0),
        panic=TempCelsius(60.0),
        threshold=None,
    )
    assert temp.get() == TempStatus(
        temp=TempCelsius(34.0),
        min=TempCelsius(40.0),
        max=TempCelsius(50.0),
        panic=TempCelsius(60.0),
        threshold=None,
        is_panic=False,
        is_threshold=False,
    )
    print(repr(temp))


def test_file_temp_min_max_files(temp_path, file_temp_path):
    with pytest.raises(RuntimeError):
        # min == max is an error
        FileTemp(
            temp_path=str(file_temp_path),
            min=None,
            max=None,
            panic=TempCelsius(60.0),
            threshold=None,
        ).get()

    temp = FileTemp(
        temp_path=str(file_temp_path),
        min=TempCelsius(50.0),
        max=None,
        panic=TempCelsius(60.0),
        threshold=None,
    )
    assert temp.get() == TempStatus(
        temp=TempCelsius(34.0),
        min=TempCelsius(50.0),
        max=TempCelsius(127.0),
        panic=TempCelsius(60.0),
        threshold=None,
        is_panic=False,
        is_threshold=False,
    )


def test_hddtemp_many(hddtemp_output_many):
    with patch.object(HDDTemp, "_call_hddtemp") as mock_call_hddtemp:
        mock_call_hddtemp.return_value = hddtemp_output_many
        t = HDDTemp(
            disk_path="/dev/sd?",
            min=TempCelsius(38.0),
            max=TempCelsius(45.0),
            panic=TempCelsius(50.0),
            threshold=None,
            hddtemp_bin="testbin",
        )

        assert t.get() == TempStatus(
            temp=TempCelsius(39.0),
            min=TempCelsius(38.0),
            max=TempCelsius(45.0),
            panic=TempCelsius(50.0),
            threshold=None,
            is_panic=False,
            is_threshold=False,
        )
        print(repr(t))


def test_hddtemp_bad(hddtemp_output_bad):
    with patch.object(HDDTemp, "_call_hddtemp") as mock_call_hddtemp:
        mock_call_hddtemp.return_value = hddtemp_output_bad
        t = HDDTemp(
            disk_path="/dev/sda",
            min=TempCelsius(38.0),
            max=TempCelsius(45.0),
            panic=TempCelsius(50.0),
            threshold=None,
            hddtemp_bin="testbin",
        )
        with pytest.raises(RuntimeError):
            t.get()


def test_hddtemp_exec_successful(temp_path):
    (temp_path / "sda").write_text("")
    (temp_path / "sdz").write_text("")
    t = HDDTemp(
        disk_path=str(temp_path / "sd") + "?",
        min=TempCelsius(38.0),
        max=TempCelsius(45.0),
        panic=TempCelsius(50.0),
        threshold=None,
        hddtemp_bin="printf '@%s'",
    )
    expected_out = "@-n@-u@C@--@{0}/sda@{0}/sdz".format(temp_path)
    assert expected_out == t._call_hddtemp()


def test_hddtemp_exec_failed():
    t = HDDTemp(
        disk_path="/dev/sd?",
        min=TempCelsius(38.0),
        max=TempCelsius(45.0),
        panic=TempCelsius(50.0),
        threshold=None,
        hddtemp_bin="false",
    )
    with pytest.raises(subprocess.CalledProcessError):
        t._call_hddtemp()


def test_command_temp_with_minmax():
    t = CommandTemp(
        shell_command=r"printf '%s\n' 35 30 40",
        min=TempCelsius(31.0),
        max=TempCelsius(39.0),
        panic=TempCelsius(50.0),
        threshold=None,
    )
    assert t.get() == TempStatus(
        temp=TempCelsius(35.0),
        min=TempCelsius(31.0),
        max=TempCelsius(39.0),
        panic=TempCelsius(50.0),
        threshold=None,
        is_panic=False,
        is_threshold=False,
    )
    print(repr(t))


def test_command_temp_without_minmax():
    t = CommandTemp(
        shell_command=r"printf '%s\n' 35 30 40",
        min=None,
        max=None,
        panic=TempCelsius(50.0),
        threshold=None,
    )
    assert t.get() == TempStatus(
        temp=TempCelsius(35.0),
        min=TempCelsius(30.0),
        max=TempCelsius(40.0),
        panic=TempCelsius(50.0),
        threshold=None,
        is_panic=False,
        is_threshold=False,
    )



from collections import defaultdict
from contextlib import ExitStack
from typing import Dict, Mapping, Optional

from afancontrol.config import (
    FanName,
    FansTempsRelation,
    MappingName,
    TempName,
    TriggerConfig,
)
from afancontrol.fans import Fans
from afancontrol.logger import logger
from afancontrol.metrics import Metrics
from afancontrol.pwmfan import PWMFanNorm, PWMValueNorm
from afancontrol.report import Report
from afancontrol.temp import Temp, TempStatus
from afancontrol.trigger import Triggers


class Manager:
    def __init__(
        self,
        *,
        fans: Mapping[FanName, PWMFanNorm],
        temps: Mapping[TempName, Temp],
        mappings: Mapping[MappingName, FansTempsRelation],
        report: Report,
        triggers_config: TriggerConfig,
        metrics: Metrics
    ) -> None:
        self.report = report
        self.fans = Fans(fans, report=report)
        self.temps = temps
        self.mappings = mappings
        self.triggers = Triggers(triggers_config, report)
        self.metrics = metrics
        self._stack = None

    def __enter__(self):  # reusable
        self._stack = ExitStack()
        try:
            self._stack.enter_context(self.fans)
            self._stack.enter_context(self.triggers)
            self._stack.enter_context(self.metrics)
        except Exception:
            self._stack.close()
            raise
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._stack.close()
        return None

    def tick(self) -> None:
        with self.metrics.measure_tick():
            temps = self._get_temps()
            self.fans.check_speeds()

            self.triggers.check(temps)

            if self.triggers.is_alerting:
                self.fans.set_all_to_full_speed()
            else:
                speeds = self._map_temps_to_fan_speeds(temps)
                self.fans.set_fan_speeds(speeds)

        try:
            self.metrics.tick(temps, self.fans, self.triggers)
        except Exception:
            logger.warning("Failed to collect metrics", exc_info=True)

    def _get_temps(self) -> Mapping[TempName, Optional[TempStatus]]:
        result = {}
        for name, temp in self.temps.items():
            try:
                status = temp.get()  # type: Optional[TempStatus]
            except Exception as e:
                status = None
                logger.warning(
                    "Temp sensor [%s] has failed: %s", name, e, exc_info=True
                )
            else:
                logger.debug("Temp status [%s]: %s", name, status)
            result[name] = status
        return result

    def _map_temps_to_fan_speeds(
        self, temps: Mapping[TempName, Optional[TempStatus]]
    ) -> Mapping[FanName, PWMValueNorm]:
        res = defaultdict(
            lambda: PWMValueNorm(0.0)
        )  # type: Dict[FanName, PWMValueNorm]

        for mapping_name, relation in self.mappings.items():
            max_speed = None  # type: Optional[PWMValueNorm]
            for temp_name in relation.temps:
                temp = temps[temp_name]
                if temp is None:
                    # Failing sensor -- this is the panic mode.
                    max_speed = PWMValueNorm(1.0)
                else:
                    speed = self._speed_for_temp_status(temp)
                    if max_speed is None:
                        max_speed = speed
                    max_speed = max(max_speed, speed)

            assert max_speed is not None

            for fan_modifier in relation.fans:
                pwm_norm = PWMValueNorm(max_speed * fan_modifier.modifier)
                pwm_norm = max(pwm_norm, PWMValueNorm(0.0))
                pwm_norm = min(pwm_norm, PWMValueNorm(1.0))
                res[fan_modifier.fan] = max(pwm_norm, res[fan_modifier.fan])

        for name in self.fans.fans.keys():
            if name not in res:
                # This fan was missing in the mapping
                res[name] = PWMValueNorm(1.0)

        return res

    def _speed_for_temp_status(self, temp: TempStatus) -> PWMValueNorm:
        speed = PWMValueNorm((temp.temp - temp.min) / (temp.max - temp.min))
        speed = max(speed, PWMValueNorm(0))
        speed = min(speed, PWMValueNorm(1))
        return speed



#!/usr/bin/env python3

import re

from setuptools import setup

with open("src/afancontrol/__init__.py", "rt") as f:
    version = re.search(r'^__version__ = "(.*?)"$', f.read()).group(1)

setup(version=version)



import abc
import sys
from time import sleep
from typing import Optional

import click

from afancontrol.arduino import (
    DEFAULT_BAUDRATE,
    ArduinoConnection,
    ArduinoName,
    ArduinoPin,
    ArduinoPWMFan,
)
from afancontrol.pwmfan import (
    BasePWMFan,
    FanInputDevice,
    FanValue,
    LinuxPWMFan,
    PWMDevice,
    PWMValue,
)

# Time to wait before measuring fan speed after setting a PWM value.
STEP_INTERVAL_SECONDS = 2

# Time to wait before starting the test right after resetting the fan
# (i.e. setting it to full speed).
FAN_RESET_INTERVAL_SECONDS = 7

EXIT_CODE_CTRL_C = 130  # https://stackoverflow.com/a/1101969

HELP_FAN_TYPE = (
    "Linux -- a standard PWM fan connected to a motherboard; "
    "Arduino -- a PWM fan connected to an Arduino board."
)

HELP_LINUX_PWM_FILE = (
    "PWM file for a Linux PWM fan, e.g. `/sys/class/hwmon/hwmon0/device/pwm2`."
)
HELP_LINUX_FAN_INPUT_FILE = (
    "Fan input (tachometer) file for a Linux PWM fan, "
    "e.g. `/sys/class/hwmon/hwmon0/device/fan2_input`."
)

HELP_ARDUINO_SERIAL_URL = "URL for the Arduino's Serial port"
HELP_ARDUINO_BAUDRATE = "Arduino Serial connection baudrate"
HELP_ARDUINO_PWM_PIN = (
    "Arduino Board pin where the target fan's PWM wire is connected to."
)
HELP_ARDUINO_TACHO_PIN = (
    "Arduino Board pin where the target fan's tachometer wire is connected to."
)

HELP_OUTPUT_FORMAT = (
    "Output format for the measurements. `csv` data could be used "
    "to make a plot using a spreadsheet program like MS Excel."
)
HELP_TEST_DIRECTION = (
    "The default test is to stop the fan and then gracefully increase its speed. "
    "You might want to reverse it, i.e. run the fan at full speed and then start "
    "decreasing the speed. This would allow you to test the fan without fully "
    "stopping it, if you abort the test with Ctrl+C when the speed becomes too low."
)
HELP_PWM_STEP_SIZE = (
    "A single step size for the PWM value. `accurate` equals to 5 and provides "
    "more accurate results, but is a slower option. `fast` equals to 25 and completes "
    "faster."
)


@click.command()
@click.option(
    "--fan-type",
    help="FAN type. %s" % HELP_FAN_TYPE,
    default="linux",
    type=click.Choice(["linux", "arduino"]),
    prompt="\n%s\nFAN type (linux, arduino)" % HELP_FAN_TYPE,
    # `show_choices` is supported since click 7.0
    show_default=True,
)
@click.option(
    "--linux-fan-pwm",
    help=HELP_LINUX_PWM_FILE,
    type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    "--linux-fan-input",
    help=HELP_LINUX_FAN_INPUT_FILE,
    type=click.Path(exists=True, dir_okay=False),
)
@click.option("--arduino-serial-url", help=HELP_ARDUINO_SERIAL_URL, type=str)
@click.option(
    "--arduino-baudrate",
    help=HELP_ARDUINO_BAUDRATE,
    type=int,
    default=DEFAULT_BAUDRATE,
    show_default=True,
)
@click.option("--arduino-pwm-pin", help=HELP_ARDUINO_PWM_PIN, type=int)
@click.option("--arduino-tacho-pin", help=HELP_ARDUINO_TACHO_PIN, type=int)
@click.option(
    "-f",
    "--output-format",
    help=HELP_OUTPUT_FORMAT,
    default="human",
    type=click.Choice(["human", "csv"]),
    prompt="\n%s\nOutput format (human, csv)" % HELP_OUTPUT_FORMAT,
    show_default=True,
)
@click.option(
    "-d",
    "--direction",
    help=HELP_TEST_DIRECTION,
    default="increase",
    type=click.Choice(["increase", "decrease"]),
    prompt="\n%s\nTest direction (increase decrease)" % HELP_TEST_DIRECTION,
    show_default=True,
)
@click.option(
    "-s",
    "--pwm-step-size",
    help=HELP_PWM_STEP_SIZE,
    default="accurate",
    type=click.Choice(["accurate", "fast"]),
    prompt="\n%s\nPWM step size (accurate fast)" % HELP_PWM_STEP_SIZE,
    show_default=True,
)
def fantest(
    *,
    fan_type: str,
    linux_fan_pwm: Optional[str],
    linux_fan_input: Optional[str],
    arduino_serial_url: Optional[str],
    arduino_baudrate: int,
    arduino_pwm_pin: Optional[int],
    arduino_tacho_pin: Optional[int],
    output_format: str,
    direction: str,
    pwm_step_size: str
) -> None:
    """The PWM fan testing program.

This program tests how changing the PWM value of a fan affects its speed.

In the beginning the fan would be stopped (by setting it to a minimum PWM value),
and then the PWM value would be increased in small steps, while also
measuring the speed as reported by the fan.

This data would help you to find the effective range of values
for the `pwm_line_start` and `pwm_line_end` settings where the correlation
between PWM and fan speed is close to linear. Usually its
`pwm_line_start = 100` and `pwm_line_end = 240`, but it is individual
for each fan. The allowed range for a PWM value is from 0 to 255.

Note that the fan would be stopped for some time during the test. If you'll
feel nervous, press Ctrl+C to stop the test and return the fan to full speed.

Before starting the test ensure that no fan control software is currently
controlling the fan you're going to test.

"""
    try:
        if fan_type == "linux":
            if not linux_fan_pwm:
                linux_fan_pwm = click.prompt(
                    "\n%s\nPWM file" % HELP_LINUX_PWM_FILE,
                    type=click.Path(exists=True, dir_okay=False),
                )

            if not linux_fan_input:
                linux_fan_input = click.prompt(
                    "\n%s\nFan input file" % HELP_LINUX_FAN_INPUT_FILE,
                    type=click.Path(exists=True, dir_okay=False),
                )

            assert linux_fan_pwm is not None
            assert linux_fan_input is not None
            fan = LinuxPWMFan(
                pwm=PWMDevice(linux_fan_pwm), fan_input=FanInputDevice(linux_fan_input)
            )  # type: BasePWMFan
        elif fan_type == "arduino":
            if not arduino_serial_url:
                arduino_serial_url = click.prompt(
                    "\n%s\nArduino Serial url" % HELP_ARDUINO_SERIAL_URL, type=str
                )

                # typeshed currently specifies `Optional[str]` for `default`,
                # see https://github.com/python/typeshed/blob/5acc22d82aa01005ea47ef64f31cad7e16e78450/third_party/2and3/click/termui.pyi#L34  # noqa
                # however the click docs say that `default` can be of any type,
                # see https://click.palletsprojects.com/en/7.x/prompts/#input-prompts
                # Hence the `type: ignore`.
                arduino_baudrate = click.prompt(  # type: ignore
                    "\n%s\nBaudrate" % HELP_ARDUINO_BAUDRATE,
                    type=int,
                    default=arduino_baudrate,
                    show_default=True,
                )
            if not arduino_pwm_pin:
                arduino_pwm_pin = click.prompt(
                    "\n%s\nArduino PWM pin" % HELP_ARDUINO_PWM_PIN, type=int
                )
            if not arduino_tacho_pin:
                arduino_tacho_pin = click.prompt(
                    "\n%s\nArduino Tachometer pin" % HELP_ARDUINO_TACHO_PIN, type=int
                )

            assert arduino_serial_url is not None
            arduino_connection = ArduinoConnection(
                name=ArduinoName("_fantest"),
                serial_url=arduino_serial_url,
                baudrate=arduino_baudrate,
            )
            assert arduino_pwm_pin is not None
            assert arduino_tacho_pin is not None
            fan = ArduinoPWMFan(
                arduino_connection,
                pwm_pin=ArduinoPin(arduino_pwm_pin),
                tacho_pin=ArduinoPin(arduino_tacho_pin),
            )
        else:
            raise AssertionError(
                "unreachable if the `fan_type`'s allowed `values` are in sync"
            )

        output = {"human": HumanMeasurementsOutput(), "csv": CSVMeasurementsOutput()}[
            output_format
        ]
        pwm_step_size_value = {"accurate": PWMValue(5), "fast": PWMValue(25)}[
            pwm_step_size
        ]
        if direction == "decrease":
            pwm_step_size_value = PWMValue(
                pwm_step_size_value * -1  # a bad PWM value, to be honest
            )
    except KeyboardInterrupt:
        click.echo("")
        sys.exit(EXIT_CODE_CTRL_C)

    try:
        run_fantest(fan=fan, pwm_step_size=pwm_step_size_value, output=output)
    except KeyboardInterrupt:
        click.echo("Fan has been returned to full speed")
        sys.exit(EXIT_CODE_CTRL_C)


def run_fantest(
    fan: BasePWMFan, pwm_step_size: PWMValue, output: "MeasurementsOutput"
) -> None:
    with fan:
        start = fan.min_pwm
        stop = fan.max_pwm
        if pwm_step_size > 0:
            print("Testing increase with step %s" % pwm_step_size)
            print("Waiting %s seconds for fan to stop..." % FAN_RESET_INTERVAL_SECONDS)
        else:
            start, stop = stop, start
            print("Testing decrease with step %s" % pwm_step_size)
            print(
                "Waiting %s seconds for fan to run in full speed..."
                % FAN_RESET_INTERVAL_SECONDS
            )

        fan.set(start)
        sleep(FAN_RESET_INTERVAL_SECONDS)

        print(output.header())

        prev_rpm = None
        for pwm_value in range(start, stop, pwm_step_size):
            fan.set(PWMValue(pwm_value))
            sleep(STEP_INTERVAL_SECONDS)
            rpm = fan.get_speed()

            rpm_delta = None  # Optional[FanValue]
            if prev_rpm is not None:
                rpm_delta = rpm - prev_rpm
            prev_rpm = rpm

            print(
                output.data_row(pwm=PWMValue(pwm_value), rpm=rpm, rpm_delta=rpm_delta)
            )

        print("Test is complete, returning fan to full speed")


class MeasurementsOutput(abc.ABC):
    @abc.abstractmethod
    def header(self) -> str:
        pass

    @abc.abstractmethod
    def data_row(
        self, pwm: PWMValue, rpm: FanValue, rpm_delta: Optional[FanValue]
    ) -> str:
        pass


class HumanMeasurementsOutput(MeasurementsOutput):
    def header(self) -> str:
        return """PWM -- PWM value;
RPM -- fan speed (as reported by the fan);
DELTA -- RPM increase since the last step."""

    def data_row(
        self, pwm: PWMValue, rpm: FanValue, rpm_delta: Optional[FanValue]
    ) -> str:
        return "PWM %s RPM %s DELTA %s" % (
            str(pwm).rjust(3),
            str(rpm).rjust(4),
            str(rpm_delta if rpm_delta is not None else "n/a").rjust(4),
        )


class CSVMeasurementsOutput(MeasurementsOutput):
    def header(self) -> str:
        return "pwm;rpm;rpm_delta"

    def data_row(
        self, pwm: PWMValue, rpm: FanValue, rpm_delta: Optional[FanValue]
    ) -> str:
        return "%s;%s;%s" % (pwm, rpm, rpm_delta if rpm_delta is not None else "")



from afancontrol.exec import exec_shell_command
from afancontrol.logger import logger


class Report:
    def __init__(self, report_command: str) -> None:
        self._report_command = report_command

    def report(self, reason: str, message: str) -> None:
        logger.info("[REPORT] Reason: %s. Message: %s", reason, message)
        try:
            rc = self._report_command
            rc = rc.replace("%REASON%", reason)
            rc = rc.replace("%MESSAGE%", message)
            exec_shell_command(rc)
        except Exception as ex:
            logger.warning("Report failed: %s", ex, exc_info=True)



import threading
from contextlib import ExitStack
from unittest.mock import patch

import pytest
from click.testing import CliRunner

from afancontrol import daemon
from afancontrol.daemon import PidFile, Signals, daemon as main


def test_main_smoke(temp_path):
    pwm_path = temp_path / "pwm" / "pwm2"
    pwm_enable_path = temp_path / "pwm" / "pwm2_enable"
    pwm_faninput_path = temp_path / "pwm" / "fan2_input"
    pwm_path.parents[0].mkdir(parents=True)
    pwm_path.write_text("100")
    pwm_enable_path.write_text("0")
    pwm_faninput_path.write_text("999")

    config_path = temp_path / "afancontrol.conf"
    config_path.write_text(
        """
[daemon]
hddtemp = true

[actions]

[temp:mobo]
type = file
path = /fake/sys/class/hwmon/hwmon0/device/temp1_input

[fan: case]
pwm = %(pwm_path)s
fan_input = %(pwm_faninput_path)s

[mapping:1]
fans = case*0.6,
temps = mobo
"""
        % dict(pwm_path=pwm_path, pwm_faninput_path=pwm_faninput_path)
    )

    with ExitStack() as stack:
        mocked_tick = stack.enter_context(patch.object(daemon.Manager, "tick"))
        stack.enter_context(patch.object(daemon, "signal"))
        stack.enter_context(
            patch.object(daemon.Signals, "wait_for_term_queued", return_value=True)
        )

        runner = CliRunner()
        result = runner.invoke(
            main,
            [
                "--verbose",
                "--config",
                str(config_path),
                "--pidfile",
                str(temp_path / "afancontrol.pid"),
                "--logfile",
                str(temp_path / "afancontrol.log"),
            ],
        )

        assert result.exit_code == 0
        assert mocked_tick.call_count == 1


def test_pidfile_not_existing(temp_path):
    pidpath = temp_path / "test.pid"
    pidfile = PidFile(str(pidpath))

    with pidfile:
        pidfile.save_pid(42)
        assert "42" == pidpath.read_text()

    assert not pidpath.exists()


def test_pidfile_existing_raises(temp_path):
    pidpath = temp_path / "test.pid"
    pidfile = PidFile(str(pidpath))
    pidpath.write_text("42")

    with pytest.raises(RuntimeError):
        with pidfile:
            pytest.fail("Should not be reached")

    assert pidpath.exists()


def test_signals():
    s = Signals()

    assert False is s.wait_for_term_queued(0.001)

    threading.Timer(0.01, lambda: s.sigterm(None, None)).start()
    assert True is s.wait_for_term_queued(1e6)
